import logging
import pathlib
import typing
from collections.abc import Iterable, Iterator

import atomlite
import numpy as np
import rdkit.Chem.AllChem as rdkit

from stk._internal.atom import Atom
from stk._internal.atom_info import AtomInfo
from stk._internal.bond import Bond
from stk._internal.bond_info import BondInfo
from stk._internal.construction_result.construction_result import (
    ConstructionResult,
)
from stk._internal.molecule import Molecule
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.utilities.molecule import (
    get_bond_info_atom_ids,
    sort_bond_atoms_by_id,
)

logger = logging.getLogger(__name__)

T = typing.TypeVar("T", bound="ConstructedMolecule")


NumBuildingBlocks: typing.TypeAlias = dict[Molecule, int]


class ConstructedMolecule(Molecule):
    """
    Represents constructed molecules.

    Examples:

        *Initialization*

        A :class:`.ConstructedMolecule` is initialized from a
        :class:`.TopologyGraph`, which is typically initialized from
        some :class:`.BuildingBlock` instances.

        .. testcode:: initialization

            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)CC=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            tetrahedron = stk.cage.FourPlusSix((bb1, bb2))
            cage = stk.ConstructedMolecule(tetrahedron)

        *Hierarchical Construction*

        A :class:`ConstructedMolecule` may be used to construct other
        :class:`ConstructedMolecule` instances, though you will
        probably have to convert it to a :class:`.BuildingBlock` first

        .. testcode:: hierarchical-construction

            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)CC=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            tetrahedron = stk.cage.FourPlusSix((bb1, bb2))
            cage = stk.ConstructedMolecule(tetrahedron)

            benzene = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('c1ccccc1'),
            )
            cage_complex = stk.host_guest.Complex(
                host=stk.BuildingBlock.init_from_molecule(cage),
                guests=benzene,
            )
            cage_complex = stk.ConstructedMolecule(cage_complex)

        Obviously, the initialization of the
        :class:`.ConstructedMolecule` depends mostly on the specifics
        of the :class:`.TopologyGraph` used, and the documentation of
        those classes should be examined for more examples.

    See Also:
        * :class:`.Atom`: Represents atoms of a constructed molecule.
        * :class:`.Bond`: Represents bonds of a constructed molecule.
        * :class:`.AtomInfo`: Holds additional information about atoms
          of a constructed molecule.
        * :class:`.BondInfo`: Holds additional information about bonds
          of a constructed molecule.

    """

    _atom_infos: tuple[AtomInfo, ...]
    _bond_infos: tuple[BondInfo, ...]
    _num_building_blocks: dict[Molecule, int]

    def __init__(self, topology_graph: TopologyGraph) -> None:
        """
        Parameters:
            topology_graph:
                The topology graph of the constructed molecule.
        """
        self._init_from_construction_result(
            obj=self,
            construction_result=topology_graph.construct(),
        )

    @classmethod
    def init(
        cls,
        atoms: Iterable[Atom],
        bonds: Iterable[Bond],
        position_matrix: np.ndarray,
        atom_infos: Iterable[AtomInfo],
        bond_infos: Iterable[BondInfo],
        num_building_blocks: "NumBuildingBlocks",
    ) -> typing.Self:
        """
        Initialize a :class:`.ConstructedMolecule` from its components.

        Parameters:

            atoms (list[Atom]):
                The atoms of the molecule.

            bonds (list[Bond]):
                The bonds of the molecule.

            position_matrix:
                A ``(n, 3)`` position matrix of the molecule.

            atom_infos (list[AtomInfo]):
                The atom infos of the molecule.

            bond_infos (list[BondInfo]):
                The bond infos of the molecule.

            num_building_blocks:
                Maps each building block of the constructed molecule to
                the number of times it is present in it.

        Returns:
            ConstructedMolecule: The constructed molecule.
        """
        molecule = cls.__new__(cls)
        Molecule.__init__(molecule, atoms, bonds, position_matrix)
        molecule._atom_infos = tuple(atom_infos)
        molecule._bond_infos = tuple(bond_infos)
        molecule._num_building_blocks = dict(num_building_blocks)
        return molecule

    @classmethod
    def init_from_construction_result(
        cls,
        construction_result: ConstructionResult,
    ) -> typing.Self:
        """
        Initialize a :class:`.ConstructedMolecule`.

        Parameters:
            construction_result:
                The result of a construction, from which the
                :class:`.ConstructedMolecule` should be initialized.
        Returns:
            ConstructedMolecule: The constructed molecule.
        """
        return cls._init_from_construction_result(
            obj=cls.__new__(cls),
            construction_result=construction_result,
        )

    @staticmethod
    def _init_from_construction_result(
        obj: T,
        construction_result: ConstructionResult,
    ) -> T:
        """
        Initialize a :class:`.ConstructedMolecule`.

        This modifies `obj`.

        Parameters:
            obj:
                The constructed molecule to initialize.

            construction_result:
                The result of a construction, from which the
                :class:`.ConstructedMolecule` should be initialized.

        Returns:
            The `obj` instance.
        """
        super(ConstructedMolecule, obj).__init__(
            atoms=construction_result.get_atoms(),
            bonds=construction_result.get_bonds(),
            position_matrix=construction_result.get_position_matrix(),
        )
        obj._atom_infos = construction_result.get_atom_infos()
        obj._bond_infos = construction_result.get_bond_infos()
        obj._num_building_blocks = {
            building_block: construction_result.get_num_building_block(
                building_block=building_block,
            )
            for building_block in construction_result.get_building_blocks()
        }
        return obj

    def clone(self) -> typing.Self:
        """
        Return a clone.

        Returns:
            ConstructedMolecule: The clone.
        """
        clone = super().clone()
        clone._atom_infos = self._atom_infos
        clone._bond_infos = self._bond_infos
        clone._num_building_blocks = dict(self._num_building_blocks)
        return clone

    def get_building_blocks(self) -> Iterator[Molecule]:
        """
        Yield the building blocks of the constructed molecule.

        Building blocks are yielded in an order based on their
        position in the constructed molecule. For two topologically
        equivalent constructed molecules, but with different building
        blocks, equivalently positioned building blocks will be
        yielded at the same time.

        Yields:
            A building block of the constructed molecule.
        """
        yield from self._num_building_blocks

    def get_num_building_block(
        self,
        building_block: Molecule,
    ) -> int:
        """
        Get the number of times `building_block` is present.

        Parameters:
            building_block:
                The building block whose frequency in the constructed
                molecule is desired.

        Returns:
            The number of times `building_block` was used in the
            construction of the constructed molecule.
        """
        return self._num_building_blocks[building_block]

    def get_atom_infos(
        self,
        atom_ids: int | Iterable[int] | None = None,
    ) -> Iterator[AtomInfo]:
        """
        Yield data about atoms in the molecule.

        Parameters:
            atom_ids (int | list[int] | None):
                The ids of atoms whose data is desired. If ``None``,
                data on all atoms will be yielded. Can be a single
                :class:`int`, if data on a single atom is desired.

        Yields:
            Data about an atom.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)

        for atom_id in atom_ids:
            yield self._atom_infos[atom_id]

    def get_bond_infos(self) -> Iterator[BondInfo]:
        """
        Yield data about bonds in the molecule.

        Yields:
            Data about a bond.
        """

        yield from self._bond_infos

    def with_canonical_atom_ordering(self) -> typing.Self:
        """
        Return a clone, with canonically ordered atoms.

        Returns:
            ConstructedMolecule: The clone.
        """
        return self.clone()._with_canonical_atom_ordering()

    def _with_canonical_atom_ordering(self) -> typing.Self:
        # Make all building blocks canonically ordered too.
        building_blocks = {
            building_block: building_block.with_canonical_atom_ordering()
            for building_block in self._num_building_blocks
        }

        # Cache these mappings for later, to avoid unnecessary
        # re-computations of canonical ordering.
        canonical_map = {
            building_block: building_block.get_canonical_atom_ids()
            for building_block in self._num_building_blocks
        }

        self._num_building_blocks = dict(
            zip(
                building_blocks.values(),
                self._num_building_blocks.values(),
            ),
        )

        ordering = rdkit.CanonicalRankAtoms(self.to_rdkit_mol())
        id_map = {
            new_id: atom.get_id()
            for new_id, atom in zip(ordering, self._atoms)
        }
        super()._with_canonical_atom_ordering()
        atom_map = {
            old_id: self._atoms[new_id]
            for old_id, new_id in enumerate(ordering)
        }
        old_atom_infos = self._atom_infos

        def get_atom_info(atom: Atom) -> AtomInfo:
            old_atom_info = old_atom_infos[id_map[atom.get_id()]]
            old_building_block = old_atom_info.get_building_block()

            if old_building_block is None:
                return AtomInfo(
                    atom=atom,
                    building_block_atom=None,
                    building_block=None,
                    building_block_id=None,
                )

            # If old_building_block is not None then neither is the
            # old_building_block_atom.
            old_building_block_atom = typing.cast(
                Atom,
                old_atom_info.get_building_block_atom(),
            )

            canonical_building_block_atom_id = canonical_map[
                old_building_block
            ][old_building_block_atom.get_id()]

            canonical_building_block = building_blocks[old_building_block]

            (canonical_building_block_atom,) = (
                canonical_building_block.get_atoms(
                    atom_ids=canonical_building_block_atom_id,
                )
            )

            return AtomInfo(
                atom=atom,
                building_block_atom=canonical_building_block_atom,
                building_block=canonical_building_block,
                building_block_id=(old_atom_info.get_building_block_id()),
            )

        def get_bond_info(info: BondInfo) -> BondInfo:
            building_block = info.get_building_block()
            return BondInfo(
                bond=sort_bond_atoms_by_id(
                    info.get_bond().with_atoms(atom_map)
                ),
                building_block=(
                    building_block
                    if building_block is None
                    else building_blocks[building_block]
                ),
                building_block_id=info.get_building_block_id(),
            )

        self._atom_infos = tuple(map(get_atom_info, self._atoms))
        self._bond_infos = tuple(
            sorted(
                map(get_bond_info, self._bond_infos),
                key=get_bond_info_atom_ids,
            )
        )
        return self

    def with_centroid(
        self,
        position: np.ndarray,
        atom_ids: int | Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Return a clone with its centroid at `position`.

        Parameters:
            position:
                This array holds the position on which the centroid of
                the clone is going to be placed.

            atom_ids (int | list[int] | None):
                The ids of atoms which should have their centroid set
                to `position`. If ``None``, all atoms are used.
        Returns:
            ConstructedMolecule: A clone with its centroid at `position`.
        """
        return super().with_centroid(position, atom_ids)

    def with_displacement(
        self,
        displacement: np.ndarray,
    ) -> typing.Self:
        """
        Return a displaced clone.

        Parameters:
            displacement:
                The displacement vector to be applied.
        Returns:
            ConstructedMolecule: A displaced clone.
        """
        return super().with_displacement(displacement)

    def with_position_matrix(
        self,
        position_matrix: np.ndarray,
    ) -> typing.Self:
        """
        Return a clone with atomic positions set by `position_matrix`.

        Parameters:
            position_matrix:
                The position matrix of the clone. The shape of the
                matrix is ``(n, 3)``.
        Returns:
            ConstructedMolecule: The clone.
        """
        return super().with_position_matrix(position_matrix)

    def with_rotation_about_axis(
        self,
        angle: float,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The clone is rotated by `angle` about `axis` on the
        `origin`.

        Parameters:
            angle:
                The size of the rotation in radians.
            axis:
                The axis about which the rotation happens. Must have
                unit magnitude.
            origin:
                The origin about which the rotation happens.
        Returns:
            ConstructedMolecule: A rotated clone.
        """
        return super().with_rotation_about_axis(angle, axis, origin)

    def with_rotation_between_vectors(
        self,
        start: np.ndarray,
        target: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The rotation is equal to a rotation from `start` to `target`.

        Given two direction vectors, `start` and `target`, this method
        applies the rotation required transform `start` to `target`
        onto the clone. The rotation occurs about the `origin`.

        For example, if the `start` and `target` vectors
        are 45 degrees apart, a 45 degree rotation will be applied to
        the clone. The rotation will be along the appropriate
        direction.

        The great thing about this method is that you as long as you
        can associate a geometric feature of the molecule with a
        vector, then the clone can be rotated so that this vector is
        aligned with `target`. The defined vector can be virtually
        anything. This means that any geometric feature of the molecule
        can be easily aligned with any arbitrary direction.

        Parameters:
            start:
                A vector which is to be rotated so that it transforms
                into the `target` vector.
            target:
                The vector onto which `start` is rotated.
            origin:
                The point about which the rotation occurs.
        Returns:
            ConstructedMolecule: A rotated clone.
        """
        return super().with_rotation_between_vectors(start, target, origin)

    def with_rotation_to_minimize_angle(
        self,
        start: np.ndarray,
        target: np.ndarray,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The clone is rotated by the rotation required to minimize
        the angle between `start` and `target`.

        Note that this function will not necessarily overlay the
        `start` and `target` vectors. This is because the possible
        rotation is restricted to the `axis`.

        Parameters:

            start:
                The vector which is rotated.

            target:
                The vector which is stationary.

            axis:
                The vector about which the rotation happens. Must have
                unit magnitude.

            origin:
                The origin about which the rotation happens.

        Returns:

            ConstructedMolecule: A rotated clone.

        Raises:

            :class:`ValueError`
                If `target` has a magnitude of 0. In this case it is
                not possible to calculate an angle between `start` and
                `target`.
        """
        return super().with_rotation_to_minimize_angle(
            start=start,
            target=target,
            axis=axis,
            origin=origin,
        )

    def with_structure_from_file(
        self,
        path: pathlib.Path | str,
        extension: str | None = None,
    ) -> typing.Self:
        """
        Return a clone, with its structure taken from a file.

        Multiple file types are supported, namely:

        #. ``.mol``, ``.sdf`` - MDL V2000 and V3000 files
        #. ``.xyz`` - XYZ files
        #. ``.mae`` - Schrodinger Maestro files
        #. ``.coord`` - Turbomole files
        #. ``.pdb`` - PDB files

        Parameters:

            path:
                The path to a molecular structure file holding updated
                coordinates for the :class:`.Molecule`.

            extension:
                If you want to treat the file as though it has a
                particular extension, put it here. Include the dot.

        Returns:

            ConstructedMolecule: A clone with atomic positions found in `path`.
        """
        return super().with_structure_from_file(str(path), extension)

    def with_structure_from_atomlite(
        self,
        database: atomlite.Database,
        key: str,
    ) -> typing.Self:
        """
        Return a clone, with its structure taken from an atomlite database.

        Parameters:

            database:
                The :mod:`atomlite` database.

            key:
                The key of the molecule in the database.

        Returns:

            Molecule: A clone with atomic positions found in `path`.

        Raises:

            ValueError: If the entry is not found in the database.

        """

        entry = database.get_entry(key)
        if entry is None:
            msg = f"Entry {key} not found in database."
            raise ValueError(msg)

        position_matrix = np.array(entry.molecule["conformers"][0])
        return self.clone().with_position_matrix(position_matrix)

    def write(
        self,
        path: pathlib.Path | str,
        atom_ids: int | Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Write the structure to a file.

        This function will write the format based on the extension
        of `path`.

        #. ``.mol``, ``.sdf`` - MDL V3000 MOL file
        #. ``.xyz`` - XYZ file
        #. ``.pdb`` - PDB file

        Parameters:
            path:
                The `path` to which the molecule should be written.
            atom_ids (int | list[int] | None):
                The atom ids of atoms to write. If ``None``,
                all atoms are used. If you use this parameter, the
                atom ids in the file may not correspond to the atom
                ids in the molecule.
        Returns:
            ConstructedMolecule: The molecule.
        """
        return super().write(path, atom_ids)



class CaseData:
    """
    A :class:`.Bond` test case.

    Attributes
    ----------
    bond : :class:`.Bond`
        The bond to test.

    atom1 : :class:`.Atom`
        The correct first atom of the bond.

    atom2 : :class:`.Atom`
        The correct second atom of the bond.

    order : :class:`int`
        The correct bond order of the bond.

    periodicity : :class:`tuple` of :class:`int`
        The correct periodicity of the bond.

    """

    def __init__(self, bond, atom1, atom2, order, periodicity):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        bond : :class:`.Bond`
            The bond to test.

        atom1 : :class:`.Atom`
            The correct first atom of the bond.

        atom2 : :class:`.Atom`
            The correct second atom of the bond.

        order : :class:`int`
            The correct bond order of the bond.

        periodicity : :class:`tuple` of :class:`int`
            The correct periodicity of the bond.

        """

        self.bond = bond
        self.atom1 = atom1
        self.atom2 = atom2
        self.order = order
        self.periodicity = periodicity



"""
Cage Vertices
=============

"""

import typing

import numpy as np
from scipy.spatial.distance import euclidean

from stk._internal.topology_graphs.vertex import Vertex
from stk._internal.utilities.utilities import (
    get_acute_vector,
    get_plane_normal,
    normalize_vector,
)

from ..utilities import _EdgeSorter, _FunctionalGroupSorter


class _CageVertex(Vertex):
    """
    Represents a vertex of a :class:`.Cage`.

    """

    def __init__(
        self,
        id: int,
        position: np.ndarray,
        use_neighbor_placement: bool = True,
        aligner_edge: int = 0,
    ) -> None:
        """
        Parameters:

            id:
                The id of the vertex.

            position:
                The position of the vertex.

            use_neighbor_placement:
                If ``True``, the position of the vertex will be updated
                based on the neighboring functional groups.

            aligner_edge:
                The edge which is used to align the :class:`.BuildingBlock`
                placed on the vertex. The first :class:`.FunctionalGroup`
                is rotated such that it lies exactly on this
                :class:`.Edge`. Must be between ``0`` and the number of
                edges the vertex is connected to.

        """

        self._use_neighbor_placement = use_neighbor_placement
        self._aligner_edge = aligner_edge
        super().__init__(id, position)

    def clone(self):
        clone = super().clone()
        clone._aligner_edge = self._aligner_edge
        clone._use_neighbor_placement = self._use_neighbor_placement
        return clone

    def _with_aligner_edge(self, aligner_edge: int) -> typing.Self:
        """
        Modify the instance.

        """

        self._aligner_edge = aligner_edge
        return self

    def with_aligner_edge(self, aligner_edge: int) -> typing.Self:
        """
        Return a clone with a different `aligner_edge`.

        Parameters:

            aligner_edge:
                The aligner edge of the clone.

        Returns:

            The clone. Has the same type as the original instance.

        """

        return self.clone()._with_aligner_edge(aligner_edge)

    def _with_use_neighbor_placement(
        self,
        use_neighbor_placement: bool,
    ) -> typing.Self:
        """
        Modify the instance.

        """

        self._use_neighbor_placement = use_neighbor_placement
        return self

    def with_use_neighbor_placement(
        self,
        use_neighbor_placement: bool,
    ) -> typing.Self:
        """
        Return a clone with a different `use_neighbor_placement`.

        Parameters:

            use_neighbor_placement:
                ``True`` if the position should be updated based on neighbors.

        Returns:

            The clone. Has the same type as the original instance.

        """

        return self.clone()._with_use_neighbor_placement(
            use_neighbor_placement
        )

    def use_neighbor_placement(self) -> bool:
        """
        ``True`` if the position should be updated based on neighbors.

        Returns:

            ``True`` if the position of the vertex should be updated
            based on the positions of functional groups on neighboring
            vertices.

        """

        return self._use_neighbor_placement

    @classmethod
    def init_at_center(
        cls,
        id: int,
        vertices: tuple[Vertex, ...],
    ) -> typing.Self:
        """
        Initialize a :class:`._CageVertex` in the middle of `vertices`.

        Parameters:

            id:
                The id of the initialized vertex.

            vertices:
                The vertices at whose center this one needs to be.

        Returns:

            The new vertex.

        """

        return cls(
            id=id,
            position=np.array(
                (
                    sum(vertex.get_position() for vertex in vertices)
                    / len(vertices)
                )
            ),
        )

    def get_aligner_edge(self) -> int:
        """
        Return the aligner edge of the vertex.

        Returns:

            The aligner edge.

        """

        return self._aligner_edge

    def __str__(self) -> str:
        return (
            f"Vertex(id={self._id}, "
            f"position={self._position.tolist()}, "
            f"aligner_edge={self._aligner_edge})"
        )


class LinearVertex(_CageVertex):
    def place_building_block(self, building_block, edges):
        assert building_block.get_num_functional_groups() == 2, (
            f"{building_block} needs to have exactly 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )

        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        fg_centroid = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )
        edge_position = edges[self._aligner_edge].get_position()
        edge_centroid = sum(edge.get_position() for edge in edges) / len(edges)
        building_block = building_block.with_rotation_between_vectors(
            start=fg_centroid - self._position,
            target=edge_position - edge_centroid,
            origin=self._position,
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        return building_block.with_rotation_to_minimize_angle(
            start=core_centroid - self._position,
            target=self._position,
            axis=normalize_vector(
                edges[0].get_position() - edges[1].get_position()
            ),
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        (fg,) = building_block.get_functional_groups(0)
        fg_position = building_block.get_centroid(fg.get_placer_ids())

        def fg_distance(edge):
            return euclidean(edge.get_position(), fg_position)

        edges = sorted(edges, key=fg_distance)
        return {fg_id: edge.get_id() for fg_id, edge in enumerate(edges)}


class NonLinearVertex(_CageVertex):
    def place_building_block(self, building_block, edges):
        assert building_block.get_num_functional_groups() > 2, (
            f"{building_block} needs to have more than 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        edge_centroid = sum(edge.get_position() for edge in edges) / len(edges)
        edge_normal = get_acute_vector(
            reference=edge_centroid,
            vector=get_plane_normal(
                points=np.array([edge.get_position() for edge in edges]),
            ),
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        placer_centroid = building_block.get_centroid(
            atom_ids=building_block.get_placer_ids(),
        )
        building_block = building_block.with_rotation_between_vectors(
            start=get_acute_vector(
                reference=core_centroid - placer_centroid,
                vector=building_block.get_plane_normal(
                    atom_ids=building_block.get_placer_ids(),
                ),
            ),
            target=edge_normal,
            origin=self._position,
        )
        fg_bonder_centroid = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )
        edge_position = edges[self._aligner_edge].get_position()
        return building_block.with_rotation_to_minimize_angle(
            start=fg_bonder_centroid - self._position,
            target=edge_position - edge_centroid,
            axis=edge_normal,
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        # The idea is to order the functional groups in building_block
        # by their angle with the vector running from the placer
        # centroid to fg0, going in the clockwise direction.
        # The edges are also ordered by their angle with the vector
        # running from the edge centroid to the aligner_edge,
        # going in the clockwise direction.
        #
        # Once the fgs and edges are ordered, zip and assign them.

        fg_sorter = _FunctionalGroupSorter(building_block)
        edge_sorter = _EdgeSorter(
            edges=edges,
            aligner_edge=edges[self._aligner_edge],
            axis=fg_sorter.get_axis(),
        )
        return {
            fg_id: edge.get_id()
            for fg_id, edge in zip(
                fg_sorter.get_items(),
                edge_sorter.get_items(),
            )
        }


class UnaligningVertex(_CageVertex):
    """
    Just places a building block, does not align.

    """

    def place_building_block(self, building_block, edges):
        return building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        return {fg_id: edge.get_id() for fg_id, edge in enumerate(edges)}

    @classmethod
    def init_at_center(cls, id, vertices):
        vertex = cls.__new__(cls)
        vertex._id = id
        vertex._position = sum(
            vertex.get_position() for vertex in vertices
        ) / len(vertices)
        vertex._use_neighbor_placement = True
        vertex._aligner_edge = 0
        return vertex


class AngledVertex(_CageVertex):
    def place_building_block(self, building_block, edges):
        assert building_block.get_num_functional_groups() == 2, (
            f"{building_block} needs to have exactly 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )

        fg_centroid = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )
        edge_position = edges[self._aligner_edge].get_position()
        edge_centroid = sum(edge.get_position() for edge in edges) / len(edges)
        building_block = building_block.with_rotation_between_vectors(
            start=fg_centroid - self._position,
            target=edge_position - edge_centroid,
            origin=self._position,
        )

        placer_centroid = building_block.get_centroid(
            atom_ids=building_block.get_placer_ids(),
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        core_to_placer = placer_centroid - core_centroid
        edge_centroid = sum(edge.get_position() for edge in edges) / len(edges)

        return building_block.with_rotation_between_vectors(
            start=core_to_placer,
            target=edge_centroid - self._position,
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        (fg,) = building_block.get_functional_groups(0)
        fg_position = building_block.get_centroid(fg.get_placer_ids())

        def fg_distance(edge):
            return euclidean(edge.get_position(), fg_position)

        edges = sorted(edges, key=fg_distance)
        return {fg_id: edge.get_id() for fg_id, edge in enumerate(edges)}



class CaseData:
    """
    A test case.

    Attributes
    ----------
    mutator : :class:`.MoleculeMutator`
        The mutator to test.

    record : :class:`.MoleculeRecord`
        The molecule to mutate.

    mutation_record : :class:`.MutationRecord`
        The correct mutation record.

    """

    def __init__(self, mutator, record, mutation_record):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        mutator : :class:`.MoleculeMutator`
            The mutator to test.

        record : :class:`.MoleculeRecord`
            The molecule to mutate.

        mutation_record : :class:`.MutationRecord`
            The correct mutation record.

        """

        self.mutator = mutator
        self.record = record
        self.mutation_record = mutation_record



import itertools as it

from ....utilities import is_equivalent_atom, normalize_ids


def test_get_atoms(case_data, get_atom_ids):
    """
    Test :meth:`.Molecule.get_atoms`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        The test case. Holds the molecule to test and the correct
        atoms.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.get_atoms`. This
        allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_atoms(
        molecule=case_data.molecule,
        atoms=case_data.atoms,
        get_atom_ids=get_atom_ids,
    )


def _test_get_atoms(molecule, atoms, get_atom_ids):
    """
    Test :meth:`.Molecule.get_atoms`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    atoms : :class:`tuple` of :class:`.Atom`
        The correct atoms of `molecule`.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.get_atoms`. This
        allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    atom_ids = get_atom_ids(molecule)
    if atom_ids is None:
        atom_ids = range(molecule.get_num_atoms())

    for atom_id, atom in it.zip_longest(
        normalize_ids(molecule, atom_ids),
        molecule.get_atoms(get_atom_ids(molecule)),
    ):
        is_equivalent_atom(atoms[atom_id], atom)



"""
XYZ Writing Utilities
=====================

"""


def _write_xyz_file(self, path, atom_ids):
    """
    Write to a ``.xyz`` file.

    This function should not be used directly, only via
    :meth:`write`.

    Parameters
    ----------
    path : :class:`str`
        The full path to the file being written.

    atom_ids : :class:`iterable` of :class:`int`
        The atom ids of atoms to write. Can be a single
        :class:`int`, if a single atom is to be used, or ``None``,
        if all atoms are to be used.

    Returns
    -------
    None : :class:`NoneType`

    """

    if atom_ids is None:
        atom_ids = range(len(self._atoms))
    elif isinstance(atom_ids, int):
        atom_ids = (atom_ids,)

    content = [0]
    for i, atom_id in enumerate(atom_ids, 1):
        x, y, z = self._position_matrix[:, atom_id]
        symbol = self._atoms[atom_id].__class__.__name__
        content.append(f"{symbol} {x:f} {y:f} {z:f}\n")
    # Set first line to the atom_count.
    content[0] = f"{i}\n\n"

    with open(path, "w") as xyz:
        xyz.write("".join(content))



import typing
from collections import abc

from stk._internal.building_block import BuildingBlock
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import (
    ReactionFactory,
)
from stk._internal.topology_graphs.edge import Edge
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import Vertex

from .vertices import SingleVertex


class InternalReaction(TopologyGraph):
    """Represents an intramolecular reaction(s).

    Here, we designed a way to take a single building block (maybe constructed
    from another series of building blocks) and perform any number of
    intramolecular reactions.

    Examples:

        *Construction*

        The number of functional groups in the core building block
        define the topology graph


        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            building_block = stk.BuildingBlock(
                smiles="CCCC(CCCCI)CCCCCC(I)CC",
                functional_groups=stk.IodoFactory(),
            )


            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        building_block.get_atoms(),
                        building_block.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in building_block.get_bonds()
                ),
            )


        .. testcode:: construction-internal

            import stk

            building_block = stk.BuildingBlock(
                smiles="CCCC(CCCCI)CCCCCC(I)CC",
                functional_groups=stk.IodoFactory(),
            )

            reacted = stk.ConstructedMolecule(stk.small.InternalReaction(
                building_block=building_block,
                num_reactions=1
            ))

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            building_block = stk.BuildingBlock(
                smiles="CCCC(CCCCI)CCCCCC(I)CC",
                functional_groups=stk.IodoFactory(),
            )

            reacted = stk.ConstructedMolecule(stk.small.InternalReaction(
                building_block=building_block,
                num_reactions=1
            ))

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        reacted.get_atoms(),
                        reacted.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in reacted.get_bonds()
                ),
            )

        And multiple reactions can be performed, where the closest pairs should
        react.

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            building_block = stk.BuildingBlock(
                smiles="ICCCCC(CCCI)CCCCCC(I)CCI",
                functional_groups=stk.IodoFactory(),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        building_block.get_atoms(),
                        building_block.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in building_block.get_bonds()
                ),
            )

        .. testcode:: construction

            import stk

            building_block = stk.BuildingBlock(
                smiles="ICCCCC(CCCI)CCCCCC(I)CCI",
                functional_groups=stk.IodoFactory(),
            )
            reacted = stk.ConstructedMolecule(stk.small.InternalReaction(
                building_block=building_block,
                num_reactions=2,
            ))


        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            building_block = stk.BuildingBlock(
                smiles="ICCCCC(CCCI)CCCCCC(I)CCI",
                functional_groups=stk.IodoFactory(),
            )

            reacted = stk.ConstructedMolecule(stk.small.InternalReaction(
                building_block=building_block,
                num_reactions=2
            ))

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        reacted.get_atoms(),
                        reacted.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in reacted.get_bonds()
                ),
            )

        *Suggested Optimization*

        For :class:`.InternalReaction` topologies, no :mod:`stk` optimizer is
        appropriate, as they all assume distinct building blocks.

    """

    def __init__(
        self,
        building_block: BuildingBlock,
        num_reactions: int,
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
        scale_multiplier: float = 1.0,
        reaction_factory: ReactionFactory = GenericReactionFactory(),
    ) -> None:
        """
        Parameters:

            building_block:
                The only building block.

            num_reactions:
                The number of internal reactions to run.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

            scale_multiplier:
                Scales the positions of the vertices.

            reaction_factory:
                The factory to use for creating reactions between
                functional groups of building blocks.

        Raises:

            :class:`ValueError`
                If the number of reactions does not match the number of pairs
                of functional groups.

        """

        self._repr = f"Internal({building_block!r}, {num_reactions!r})"
        self._num_reactions = num_reactions
        num_functional_groups = building_block.get_num_functional_groups()
        if num_functional_groups != 2 * self._num_reactions:
            raise ValueError(
                f"The number of reactions {num_reactions} does match the number"
                f" of pairs of functional groups ({num_functional_groups})."
            )

        vertices = (SingleVertex(0, (0, 0, 0)),)
        edges = []
        for potential_edge in range(self._num_reactions):
            edges.append(
                Edge(
                    id=potential_edge,
                    vertex1=vertices[0],
                    vertex2=vertices[0],
                )
            )

        super().__init__(
            building_block_vertices={building_block: vertices},
            edges=tuple(edges),
            reaction_factory=reaction_factory,
            construction_stages=(),
            optimizer=optimizer,
            num_processes=num_processes,
            scale_multiplier=scale_multiplier,
        )

    def clone(self) -> typing.Self:
        clone = self._clone()
        clone._repr = self._repr
        clone._num_reactions = self._num_reactions
        return clone

    @staticmethod
    def _get_scale(
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
        scale_multiplier: float,
    ) -> float:
        return 1

    def with_building_blocks(
        self,
        building_block_map: dict[BuildingBlock, BuildingBlock],
    ) -> typing.Self:
        return self.clone()._with_building_blocks(building_block_map)

    def __repr__(self) -> str:
        return self._repr



import itertools as it


def test_get_deleters(generic_case_data):
    """
    Test :meth:`.GenericFunctionalGroup.get_deleters`.

    Parameters
    ----------
    generic_case_data : :class:`.GenericCaseData`
        The test case. Holds the functional group to test and the
        correct deleter atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_deleters(
        functional_group=generic_case_data.functional_group,
        deleters=generic_case_data.deleters,
    )


def _test_get_deleters(functional_group, deleters):
    """
    Test :meth:`.GenericFunctionalGroup.get_deleters`.

    Parameters
    ----------
    functional_group : :class:`.GenericFunctionalGroup`
        The functional group to test.

    deleters : :class:`tuple` of :class:`.Atom`
        The correct deleter atoms.

    Returns
    -------
    None : :class:`NoneType`

    """

    for atom1, atom2 in it.zip_longest(
        functional_group.get_deleters(),
        deleters,
    ):
        assert atom1 is atom2



def test_get_atom2(case_data):
    """
    Test :meth:`.Bond.get_atom2`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the bond to test and the correct *atom2*.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.bond.get_atom2() is case_data.atom2



from dataclasses import dataclass
from typing import Callable

import pymongo
import pytest
import rdkit.Chem.AllChem as rdkit
import stk

from ..case_data import CaseData


@dataclass(frozen=True)
class CaseDataData:
    """
    Data used to create a :class:`.CaseDataData` instance.

    Attributes
    ----------
    get_database : :class:`callable`
        Creates the database to test. Takes a
        :class:`pymongo.MongoClient` as input and returns a
        :class:`.MoleculeMongoDb` instance.

    molecule : :class:`.Molecule`
        The molecule to put and get from the database being tested.

    key : :class:`object`
        The key used to retrieve the :attr:`.molecule` from the
        database.

    """

    get_database: Callable[[pymongo.MongoClient], stk.ValueMongoDb]
    molecule: stk.Molecule
    key: object


@pytest.fixture(
    params=(
        lambda: CaseDataData(
            get_database=lambda mongo_client: stk.MoleculeMongoDb(
                mongo_client=mongo_client,
                database="_stk_test_database_for_testing",
                put_lru_cache_size=0,
                get_lru_cache_size=0,
            ),
            molecule=stk.BuildingBlock("BrCCBr"),
            key={
                "InChIKey": rdkit.MolToInchiKey(rdkit.MolFromSmiles("BrCCBr"))
            },
        ),
        lambda: CaseDataData(
            get_database=lambda mongo_client: stk.MoleculeMongoDb(
                mongo_client=mongo_client,
                database="_stk_test_database_for_testing",
                put_lru_cache_size=0,
                get_lru_cache_size=0,
                jsonizer=stk.MoleculeJsonizer(
                    key_makers=(
                        stk.MoleculeKeyMaker(
                            key_name="SMILES",
                            get_key=lambda molecule: rdkit.MolToSmiles(
                                molecule.to_rdkit_mol()
                            ),
                        ),
                    ),
                ),
            ),
            molecule=stk.BuildingBlock("BrBr"),
            key={"SMILES": "BrBr"},
        ),
        lambda: CaseDataData(
            get_database=lambda mongo_client: stk.MoleculeMongoDb(
                mongo_client=mongo_client,
                database="_stk_test_database_for_testing",
                put_lru_cache_size=128,
                get_lru_cache_size=128,
                jsonizer=stk.MoleculeJsonizer(
                    key_makers=(
                        stk.MoleculeKeyMaker(
                            key_name="SMILES",
                            get_key=lambda molecule: rdkit.MolToSmiles(
                                molecule.to_rdkit_mol()
                            ),
                        ),
                    ),
                ),
            ),
            molecule=stk.BuildingBlock("BrBr"),
            key={"SMILES": "BrBr"},
        ),
    ),
)
def molecule_mongo_db(
    request,
    mongo_client: pymongo.MongoClient,
) -> CaseData:
    data = request.param()
    return CaseData(
        database=data.get_database(mongo_client),
        molecule=data.molecule,
        key=data.key,
    )



import itertools as it

import numpy as np
import stk


def is_clone(construction_state, clone):
    atoms = it.zip_longest(
        construction_state.get_atoms(),
        clone.get_atoms(),
    )
    for atom1, atom2 in atoms:
        assert atom1 is atom2

    bonds = it.zip_longest(
        construction_state.get_bonds(),
        clone.get_bonds(),
    )
    for bond1, bond2 in bonds:
        assert bond1 is bond2

    atom_infos = it.zip_longest(
        construction_state.get_atom_infos(),
        clone.get_atom_infos(),
    )
    for atom_info1, atom_info2 in atom_infos:
        assert atom_info1 is atom_info2

    bond_infos = it.zip_longest(
        construction_state.get_bond_infos(),
        clone.get_bond_infos(),
    )
    for bond_info1, bond_info2 in bond_infos:
        assert bond_info1 is bond_info2

    assert construction_state.get_num_vertices() == clone.get_num_vertices()

    for vertex_id in range(clone.get_num_vertices()):
        assert construction_state.get_vertex(vertex_id) is clone.get_vertex(
            vertex_id
        )

    assert np.all(
        np.equal(
            construction_state.get_position_matrix(),
            clone.get_position_matrix(),
        )
    )

    assert construction_state.get_num_edges() == clone.get_num_edges()
    for edge_id in range(clone.get_num_edges()):
        assert construction_state.get_edge(edge_id) is clone.get_edge(edge_id)

    edge_group = stk.EdgeGroup(
        edges=map(clone.get_edge, range(clone.get_num_edges())),
    )
    functional_groups = it.zip_longest(
        construction_state.get_edge_group_functional_groups(
            edge_group=edge_group,
        ),
        clone.get_edge_group_functional_groups(edge_group),
    )
    for fg1, fg2 in functional_groups:
        assert fg1 is fg2

    counts1 = construction_state.get_building_block_counts()
    counts2 = clone.get_building_block_counts()
    assert (counts1.keys() | counts2.keys()) == counts1.keys()
    for building_block, count1 in counts1.items():
        assert count1 == counts2[building_block]

    assert all(
        np.all(np.equal(actual_constant, expected_constant))
        for actual_constant, expected_constant in it.zip_longest(
            construction_state.get_lattice_constants(),
            clone.get_lattice_constants(),
        )
    )



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.TwoPlusFour(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1C2(Br)[C+]=N[C+]2[C+](Br)[C+]("
                                "Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            ),
            smiles=(
                "[C+]1=NC2=C1C13[C+]=N[C+]1[C+]1C4=C(N=[C+]4)[C+]4[C+]"
                "2[C+2][C+]2C5=C([C+]=N5)[C+]3[C+2][C+]1C1=C(N=[C+]1)C"
                "21[C+]=N[C+]41"
            ),
            name=name,
        ),
    ),
)
def cage_two_plus_four(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import itertools as it

import numpy as np

from .utilities import is_equivalent_atom


def test_get_new_atoms(case_data):
    """
    Test that correct atoms are added by a :class:`.Reaction`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Includes the reaction to be tested and the atoms,
        which should be added.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_new_atoms(
        reaction_result=case_data.reaction.get_result(),
        new_atoms=case_data.new_atoms,
    )


def _test_get_new_atoms(reaction_result, new_atoms):
    """
    Test that the correct atoms are added by `reaction_result`.

    Parameters
    ----------
    reaction_result : :class:`.ReactionResult`
        The result of a :class:`.Reaction`.

    new_atoms : :class:`tuple` of :class:`.NewAtom`
        The atoms, which should be added.

    Returns
    -------
    None : :class:`NoneType`

    """

    for (atom1, position1), (atom2, position2) in it.zip_longest(
        reaction_result.get_new_atoms(),
        new_atoms,
    ):
        is_equivalent_atom(atom1, atom2)
        assert np.allclose(position1, position1, atol=1e-32)



"""
Diol Factory
============

"""

from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.utilities import get_atom_ids
from stk._internal.functional_groups.diol import Diol


class DiolFactory(FunctionalGroupFactory):
    """
    Creates :class:`.Diol` instances.

    Creates functional groups from substructures, which match the
    ``[H][O][#6]~[#6][O][H]`` functional group string.

    Examples
    --------
    *Creating Functional Groups with the Factory*

    You want to create a building block which has :class:`.Diol`
    functional groups. You want the carbon atoms in those functional
    groups to be the *bonder* atoms, and the OH groups to be a leaving
    groups.

    .. testcode:: creating-functional-groups-with-the-factory

        import stk

        building_block = stk.BuildingBlock(
            smiles='CCCC(O)C(O)CCCC',
            functional_groups=(stk.DiolFactory(), ),
        )

    .. testcode:: creating-functional-groups-with-the-factory
        :hide:

        assert all(
            isinstance(functional_group, stk.Diol)
            for functional_group
            in building_block.get_functional_groups()
        )
        assert building_block.get_num_functional_groups() == 1

    *Changing the Bonder and Deleter Atoms*

    You want to create a building block which has :class:`.Diol`
    functional groups. You want the oxygen atoms to be the *bonder*
    atoms and the hydrogen atoms to be the *deleter* atoms.

    .. testcode:: changing-the-bonder-and-deleter-atoms

        import stk

        diol_factory = stk.DiolFactory(
            # The indices of the oxygen atoms in the functional
            # group string (see docstring) are 1 and 4.
            bonders=(1, 4),
            # The indices of the hydrogen atoms in the functional
            # group string (see docstring) are 0 and 5.
            deleters=(0, 5),
        )
        building_block = stk.BuildingBlock(
            smiles='CCCC(O)C(O)CCCC',
            functional_groups=(diol_factory, ),
        )

    .. testcode:: changing-the-bonder-and-deleter-atoms
        :hide:

        fg, = building_block.get_functional_groups()
        assert fg.get_num_bonders() == 2
        assert sum(1 for _ in fg.get_deleters()) == 2

        assert all(
            isinstance(atom, stk.O)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_bonders()
        )
        assert all(
            isinstance(atom, stk.H)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_deleters()
        )

    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(
        self,
        bonders=(2, 3),
        deleters=(0, 1, 4, 5),
        placers=None,
    ):
        """
        Initialize a :class:`.DiolFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        ids = get_atom_ids("[H][O][#6]~[#6][O][H]", molecule)
        for atom_ids in ids:
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield Diol(
                hydrogen1=atoms[0],
                oxygen1=atoms[1],
                atom1=atoms[2],
                atom2=atoms[3],
                oxygen2=atoms[4],
                hydrogen2=atoms[5],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



"""
Mol Writer
==========

"""

from __future__ import annotations

import pathlib
import typing

from stk._internal.molecule import Molecule
from stk._internal.utilities.utilities import OneOrMany


class MolWriter:
    """
    A writer class for V3000 ``.mol`` files.

    Examples:

        *Writing to a File*

        .. testcode:: writing-to-a-file

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])

            writer = stk.MolWriter()
            writer.write(molecule=bb1, path='bb1.mol')

        .. testcode:: writing-to-a-file
            :hide:

            import os

            assert os.path.exists('bb1.mol')

        .. testcleanup:: writing-to-a-file

            os.remove('bb1.mol')

    """

    def _write_content(
        self,
        molecule: Molecule,
        atom_ids: typing.Optional[OneOrMany[int]],
    ) -> str:
        if atom_ids is None:
            atom_ids = range(molecule.get_num_atoms())
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)

        atom_lines = []
        coords = molecule.get_position_matrix()
        # This set gets used by bonds.
        id_map = {}
        for new_atom_id, old_atom_id in enumerate(atom_ids, 1):
            id_map[old_atom_id] = new_atom_id

            x, y, z = (i for i in coords[old_atom_id])
            (atom,) = molecule.get_atoms(atom_ids=old_atom_id)
            symbol = atom.__class__.__name__
            charge_ = atom.get_charge()
            charge = f" CHG={charge_}" if charge_ else ""
            atom_lines.append(
                f"M  V30 {new_atom_id} {symbol} {x:.4f} "
                f"{y:.4f} {z:.4f} 0{charge}\n"
            )
        atom_block = "".join(atom_lines)

        bond_lines: list[str] = []
        for bond in molecule.get_bonds():
            a1 = bond.get_atom1().get_id()
            a2 = bond.get_atom2().get_id()
            if a1 in id_map and a2 in id_map:
                bond_lines.append(
                    f"M  V30 {len(bond_lines) + 1} "
                    f"{int(bond.get_order())} "
                    f"{id_map[a1]} {id_map[a2]}\n"
                )

        num_bonds = len(bond_lines)
        bond_block = "".join(bond_lines)
        return (
            "\n"
            "     RDKit          3D\n"
            "\n"
            "  0  0  0  0  0  0  0  0  0  0999 V3000\n"
            "M  V30 BEGIN CTAB\n"
            f"M  V30 COUNTS {len(id_map)} {num_bonds} 0 0 0\n"
            "M  V30 BEGIN ATOM\n"
            f"{atom_block}"
            "M  V30 END ATOM\n"
            "M  V30 BEGIN BOND\n"
            f"{bond_block}"
            "M  V30 END BOND\n"
            "M  V30 END CTAB\n"
            "M  END\n"
            "\n"
            "$$$$\n"
        )

    def to_string(
        self,
        molecule: Molecule,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
    ) -> str:
        """
        Get a V3000 ``.mol`` file format string of `molecule`.

        Parameters:

            molecule:
                Molecule to write to V3000 ``.mol`` format.

            atom_ids:
                The atom ids of atoms to write. Can be a single
                :class:`int`, if a single atom is to be used, or
                ``None``, if all atoms are to be used.

        Returns:

            String in V3000 ``.mol`` file format.

        """

        return self._write_content(molecule, atom_ids)

    def write(
        self,
        molecule: Molecule,
        path: pathlib.Path | str,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
    ) -> None:
        """
        Write `molecule` to V3000 ``.mol`` file format.

        Parameters:

            molecule:
                Molecule to write to V3000 ``.mol`` format.

            path:
                The full path to the file being written.

            atom_ids:
                The atom ids of atoms to write. Can be a single
                :class:`int`, if a single atom is to be used, or
                ``None``, if all atoms are to be used.

        """

        with open(path, "w") as f:
            f.write(self._write_content(molecule, atom_ids))



class CaseData:
    """
    A test case.

    Attributes
    ----------
    fitness_calculator : :class:`.FitnessCalculator`
        The fitness calculator to test.

    molecule : :class:`.Molecule`
        The molecule whose fitness is calculated.

    fitness_value : :class:`object`
        The correct fitness value.

    """

    def __init__(self, fitness_calculator, molecule, fitness_value):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        fitness_calculator : :class:`.FitnessCalculator`
            The fitness calculator to test.

        molecule : :class:`.Molecule`
            The molecule whose fitness is calculated.

        fitness_value : :class:`object`
            The correct fitness value.

        """

        self.fitness_calculator = fitness_calculator
        self.molecule = molecule
        self.fitness_value = fitness_value



"""
Iodo
====

"""

from .generic_functional_group import GenericFunctionalGroup


class Iodo(GenericFunctionalGroup):
    """
    Represents an iodo functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[iodine][atom]``.

    """

    def __init__(self, iodine, atom, bonders, deleters, placers=None):
        """
        Initialize a :class:`.Iodo` instance.

        Parameters
        ----------
        iodine : :class:`.I`
            The ``[iodine]`` atom.

        atom : :class:`.Atom`
            The ``[atom]`` atom.

        bonders : :class:`tuple` of :class:`.Atom`
            The bonder atoms.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms.

        placers : :class:`tuple` of :class:`.Atom`, optional
            The placer atoms. If ``None`` the `bonders` will be used.

        """

        self._iodine = iodine
        self._atom = atom
        super().__init__(
            atoms=(iodine, atom),
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_iodine(self):
        """
        Get the ``[iodine]`` atom.

        Returns
        -------
        :class:`.I`
            The ``[iodine]`` atom.

        """

        return self._iodine

    def get_atom(self):
        """
        Get the ``[atom]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom]`` atom.

        """

        return self._atom

    def clone(self):
        clone = super().clone()
        clone._iodine = self._iodine
        clone._atom = self._atom
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._iodine = atom_map.get(
            self._iodine.get_id(),
            self._iodine,
        )
        clone._atom = atom_map.get(
            self._atom.get_id(),
            self._atom,
        )
        return clone

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self._iodine}, {self._atom}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



import typing

from stk._internal.ea.mutation.record import MutationRecord

T = typing.TypeVar("T")


class MoleculeMutator(typing.Protocol[T]):
    """
    Performs mutation operations.

    Examples:

        *Implementation mutation operations*

        You only need to implement :meth:`.mutate`. The source code of any
        of the classes listed in :mod:`.mutator` can serve as good
        examples.

    """

    def mutate(self, record: T) -> MutationRecord[T] | None:
        """
        Return a mutant of `record`.

        Parameters:
            record:
                The molecule to be mutated.

        Returns:
            A record of the mutation or ``None``
            if `record` cannot be mutated.
        """
        ...



import typing
from collections import Counter
from collections.abc import Iterable, Iterator

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.key_makers.molecule import MoleculeKeyMaker

BatchKey: typing.TypeAlias = frozenset[tuple[str, int]]

T = typing.TypeVar("T", bound=MoleculeRecord)


class Batch(typing.Generic[T]):
    """
    Represents a batch of molecule records.

    Batches can be compared, the comparison is based on their
    fitness values. Batches can also be iterated through, this
    iterates through all the records in the batch.

    Examples:

        *Sorting Batches by Fitness Value*

        Sorting batches causes them to be sorted by fitness value.

        .. testcode:: sorting-batches-by-fitness-value

            import stk

            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[
                        stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                    ],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[
                        stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                    ],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record3 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[
                        stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                    ],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )

            batches = (
                stk.Batch(
                    records={record1: 1},
                    key_maker=stk.Inchi(),
                ),
                stk.Batch(
                    records={record2: 2},
                    key_maker=stk.Inchi(),
                ),
                stk.Batch(
                    records={record3: 3},
                    key_maker=stk.Inchi(),
                ),
            )
            sorted_batches = sorted(batches)

        .. testcode:: sorting-batches-by-fitness-value
            :hide:

            assert (
                sorted_batches[0] < sorted_batches[1] < sorted_batches[2]
            )

        *Comparing Batches by Fitness Value*

        Comparison is also based on fitness value

        .. testcode:: comparing-batches-by-fitness-value

            import stk

            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[
                        stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                    ],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            batch1 = stk.Batch(
                records={record1: 1},
                key_maker=stk.Inchi(),
            )

            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[
                        stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                    ],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            batch2 = stk.Batch(
                records={record2: 2},
                key_maker=stk.Inchi(),
            )

            if batch1 < batch2:
                print('batch1 has a smaller fitness value than batch2.')

        .. testoutput:: comparing-batches-by-fitness-value
            :hide:

            batch1 has a smaller fitness value than batch2.

        *Iterating Through Molecule Records in a Batch*

        Batches can be iterated through to get the molecule records in the
        batch

        .. testcode:: iterating-through-molecule-records-in-a-batch

            import stk

            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[
                        stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                    ],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[
                        stk.BuildingBlock('BrCCBr', stk.BromoFactory()),
                    ],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            batch = stk.Batch(
                records={record1: 1, record2: 2},
                key_maker=stk.Inchi(),
            )
            for record in batch:
                # Do stuff with record.
                pass
    """

    __slots__ = "_records", "_fitness_value", "_identity_key"

    def __init__(
        self,
        records: dict[T, float] | Iterable[tuple[T, float]],
        key_maker: MoleculeKeyMaker,
    ) -> None:
        """
        Parameters:
            records (dict[T, float]):
                The records which form the batch, mapped to their
                fitness values.
            key_maker:
                Used to make keys for molecules, which are used to
                determine the identity key of the batch. If two
                batches have the same molecule keys, the same number of
                times, they will have the same identity key.
        """
        self._records = dict(records)
        self._fitness_value = sum(self._records.values())
        molecules = (record.get_molecule() for record in self._records)
        self._identity_key = frozenset(
            Counter(map(key_maker.get_key, molecules)).items()
        )

    def get_size(self) -> int:
        """
        Get the number of molecules in the batch.

        Returns:
            The number of molecules in the batch.
        """
        return len(self._records)

    def get_fitness_value(self) -> float:
        """
        Get the fitness value of the batch.

        Returns:
            The fitness value.
        """
        return self._fitness_value

    def get_identity_key(self) -> BatchKey:
        """
        Get the identity key of the batch.

        If two batches hold the same molecules, the same number of
        times, they will have the same identity key.

        Returns:
            A hashable object which can be used to compare if two
            batches have the same identity.
        """
        return self._identity_key

    def __iter__(self) -> Iterator[T]:
        return iter(self._records)

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, Batch):
            return NotImplemented
        return self._fitness_value == other._fitness_value

    def __gt__(self, other: "Batch[T]") -> bool:
        return self._fitness_value > other._fitness_value

    def __ge__(self, other: "Batch[T]") -> bool:
        return self._fitness_value >= other._fitness_value

    def __lt__(self, other: "Batch[T]") -> bool:
        return self._fitness_value < other._fitness_value

    def __le__(self, other: "Batch[T]") -> bool:
        return self._fitness_value <= other._fitness_value

    def __repr__(self) -> str:
        return f"Batch({self._fitness_value})"

    def __str__(self) -> str:
        return repr(self)



class CaseData:
    """
    A :class:`.ReactionFactory` test case.

    Attributes
    ----------
    factory : :class:`.ReactionFactory`
        The reaction factory to test.

    construction_state : :class:`.ConstructionState`
        The construction state to pass to the factory.

    edge_group : :class:`.EdgeGroup`
        The edge group to pass to the factory.

    reaction_result : :class:`.ReactionResult`
        The expected result of the reaction returned by the factory.

    Notes
    -----
    The reason the test case looks at the :attr:`.reaction_result`,
    rather than the actual :class:`.Reaction` returned by the
    :attr:`factory`, is that :class:`.Reaction` does not provide a
    public API through which its identity can be determined, except for
    the :attr:`.reaction_result`.

    """

    def __init__(
        self,
        factory,
        construction_state,
        edge_group,
        reaction_result,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        factory : :class:`.ReactionFactory`
            The reaction factory to test.

        construction_state : :class:`.ConstructionState`
            The construction state to pass to the factory.

        edge_group : :class:`.EdgeGroup`
            The edge group to pass to the factory.

        reaction_result : :class:`.ReactionResult`
            The expected result of the reaction returned by the
            factory.

        """

        self.factory = factory
        self.construction_state = construction_state
        self.edge_group = edge_group
        self.reaction_result = reaction_result



class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    atoms : :class:`tuple`
        The correct atoms of the molecule.

    """

    def __init__(self, molecule, atoms):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule to test.

        atoms : :class:`.Atom`
            The correct atoms of the molecule.

        """

        self.molecule = molecule
        self.atoms = atoms



from typing import Any

import pytest
import stk

from ..case_data import CaseData


def get_topology_graph(num_repeating_units: int) -> stk.TopologyGraph:
    return stk.polymer.Linear(
        building_blocks=[
            stk.BuildingBlock("BrCCBr", stk.BromoFactory()),
        ],
        repeating_unit="A",
        num_repeating_units=num_repeating_units,
    )


@pytest.fixture(scope="session")
def stochastic_universal_sampling_population_1() -> dict[
    stk.MoleculeRecord[Any], float
]:
    return {
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(2),
        ): 10,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(3),
        ): 9,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(4),
        ): 2,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(5),
        ): 1,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(6),
        ): 1,
    }


@pytest.fixture(
    scope="session",
    params=(
        lambda population: CaseData.new(
            selector=stk.StochasticUniversalSampling(
                duplicate_molecules=False,
            ),
            population=population,
            selected=([0], [1], [2], [3], [4]),
        ),
        lambda population: CaseData.new(
            selector=stk.StochasticUniversalSampling(
                duplicate_batches=False,
            ),
            population=population,
            selected=([0], [1], [2], [3], [4]),
        ),
        lambda population: CaseData.new(
            selector=stk.StochasticUniversalSampling(
                batch_size=2, duplicate_batches=False
            ),
            population=population,
            selected=(
                [0, 1],
                [0, 2],
                [0, 3],
                [0, 4],
                [1, 2],
                [1, 3],
                [1, 4],
                [2, 3],
                [2, 4],
                [3, 4],
            ),
        ),
    ),
)
def stochastic_universal_sampling(
    request,
    stochastic_universal_sampling_population_1: dict[
        stk.MoleculeRecord[Any], float
    ],
):
    return request.param(stochastic_universal_sampling_population_1)



"""
Periodic Info
=============

Class holding periodic cell information.

"""

from __future__ import annotations

import logging

import numpy as np

from stk._internal.utilities.utilities import cap_absolute_value

logger = logging.getLogger(__name__)


class PeriodicInfo:
    """
    Periodic cell information for periodic systems.

    """

    def __init__(
        self,
        vector_1: np.ndarray,
        vector_2: np.ndarray,
        vector_3: np.ndarray,
    ) -> None:
        """
        Initialize a :class:`.PeriodicInfo` instance.

        Converts cell matrix to lengths and angles, where lengths are
        in Angstrom and angles are in degrees. This code is modified
        from the pymatgen source code [1]_.

        Parameters:

            vector_1:
                First cell lattice vector of shape (3, ) in Angstrom.

            vector_2:
                Second cell lattice vector of shape (3, ) in Angstrom.

            vector_3:
                Third cell lattice vector of shape (3, ) in Angstrom.

        References:

            .. [1] https://pymatgen.org/_modules/pymatgen/core/\
lattice.html

        """

        self._vector_1 = np.array(vector_1)
        self._vector_2 = np.array(vector_2)
        self._vector_3 = np.array(vector_3)
        self._cell_matrix = (
            self._vector_1,
            self._vector_2,
            self._vector_3,
        )

        a, b, c = tuple(
            np.sqrt(np.sum(i**2)).tolist() for i in self._cell_matrix
        )
        self._a = a
        self._b = b
        self._c = c

        lengths = (a, b, c)
        angles = np.zeros(3)
        for i in range(3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            angles[i] = cap_absolute_value(
                value=(
                    np.dot(self._cell_matrix[j], self._cell_matrix[k])
                    / (lengths[j] * lengths[k])
                ),
            )
        angles = np.arccos(angles) * 180.0 / np.pi

        alpha, beta, gamma = angles
        self._alpha = alpha
        self._beta = beta
        self._gamma = gamma

    def clone(self) -> PeriodicInfo:
        """
        Return a clone.

        Returns:

            The clone.

        """

        clone = self.__class__.__new__(self.__class__)
        PeriodicInfo.__init__(
            self=clone,
            vector_1=self._vector_1,
            vector_2=self._vector_2,
            vector_3=self._vector_3,
        )
        return clone

    def get_vector_1(self) -> np.ndarray:
        """
        Get *x* vector.

        Returns:

            Cell lattice vector of shape (3, ) in *x* direction in
            Angstrom.

        """

        return np.array(self._vector_1)

    def get_vector_2(self) -> np.ndarray:
        """
        Get *y* vector.

        Returns:

            Cell lattice vector of shape (3, ) in *y* direction in
            Angstrom.

        """

        return np.array(self._vector_2)

    def get_vector_3(self) -> np.ndarray:
        """
        Get *z* vector.

        Returns:

            Cell lattice vector of shape (3, ) in *z* direction in
            Angstrom.

        """

        return np.array(self._vector_3)

    def get_cell_matrix(
        self,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Get cell matrix.

        Returns:

            Tuple of length three containing *x*, *y* and *z* direction
            lattice vector of shape (3, ) in Angstrom.

        """

        a, b, c = self._cell_matrix
        return (np.array(a), np.array(b), np.array(c))

    def get_a(self) -> float:
        """
        Get *a* length.

        Returns:

            Length of cell along *a* direction in Angstrom.

        """

        return self._a

    def get_b(self) -> float:
        """
        Get *b* length.

        Returns:

            Length of cell along *b* direction in Angstrom.

        """

        return self._b

    def get_c(self) -> float:
        """
        Get *c* length.

        Returns:

            Length of cell along *c* direction in Angstrom.

        """

        return self._c

    def get_alpha(self) -> float:
        """
        Get *alpha* angle.

        Returns:

            *Alpha* angle of cell in degrees.

        """

        return self._alpha

    def get_beta(self) -> float:
        """
        Get *beta* angle.

        Returns:

            *Beta* angle of cell in degrees.

        """

        return self._beta

    def get_gamma(self) -> float:
        """
        Get *gamma* angle.

        Returns:

            *Gamma* angle of cell in degrees.

        """

        return self._gamma

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}(a={self._a}, b={self._b}, "
            f"c={self._c}, alpha={self._alpha}, beta={self._beta}, "
            f"gamma={self._gamma})"
        )

    def __repr__(self) -> str:
        return str(self)



class CaseData:
    """
    A test case.

    Attributes
    ----------
    start : :class:`numpy.ndarray`
        The starting position of the tested edge.

    scale : :class:`float` or :class:`numpy.ndarray`
        The scale applied to the tested edge.

    target : :class:`numpy.ndarray`
        The expected position of the new edge.

    """

    def __init__(self, start, scale, target):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        start : :class:`numpy.ndarray`
            The starting position of the tested edge.

        scale : :class:`float` or :class:`numpy.ndarray`
            The scale applied to the tested edge.

        target : :class:`numpy.ndarray`
            The expected position of the new edge.

        """

        self.start = start
        self.scale = scale
        self.target = target



import pytest
import stk
from pytest_lazyfixture import lazy_fixture

# Fixtures must be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    scope="session",
    params=(
        lambda: stk.Alcohol(
            oxygen=stk.O(0),
            hydrogen=stk.H(1),
            atom=stk.C(2),
            bonders=(stk.C(2),),
            deleters=(stk.O(0), stk.H(1)),
        ),
        lambda: stk.Aldehyde(
            carbon=stk.C(0),
            oxygen=stk.O(1),
            hydrogen=stk.H(2),
            atom=stk.C(3),
            bonders=(stk.C(0),),
            deleters=(stk.O(1),),
        ),
        lambda: stk.Alkene(
            carbon1=stk.C(0),
            atom1=stk.H(1),
            atom2=stk.H(2),
            carbon2=stk.C(3),
            atom3=stk.C(4),
            atom4=stk.H(5),
            bonders=(stk.C(4),),
            deleters=(stk.C(0), stk.H(1), stk.H(2)),
        ),
        lambda: stk.Alkyne(
            carbon1=stk.C(0),
            atom1=stk.H(1),
            carbon2=stk.C(2),
            atom2=stk.C(3),
            bonders=(stk.C(2),),
            deleters=(stk.C(0), stk.H(1)),
        ),
        lambda: stk.Amide(
            carbon=stk.C(0),
            oxygen=stk.O(1),
            nitrogen=stk.N(2),
            hydrogen1=stk.H(3),
            hydrogen2=stk.H(4),
            atom=stk.C(5),
            bonders=(stk.C(0),),
            deleters=(stk.N(2), stk.H(3), stk.H(4)),
        ),
        lambda: stk.BoronicAcid(
            boron=stk.B(0),
            oxygen1=stk.O(1),
            hydrogen1=stk.H(2),
            oxygen2=stk.O(3),
            hydrogen2=stk.H(4),
            atom=stk.C(5),
            bonders=(stk.B(0),),
            deleters=(stk.O(1), stk.H(2)),
        ),
        lambda: stk.Bromo(
            bromine=stk.Br(0),
            atom=stk.C(1),
            bonders=(stk.C(1),),
            deleters=(stk.Br(0),),
        ),
        lambda: stk.CarboxylicAcid(
            carbon=stk.C(0),
            oxygen1=stk.O(1),
            oxygen2=stk.O(2),
            hydrogen=stk.H(3),
            atom=stk.C(4),
            bonders=(stk.C(0),),
            deleters=(stk.O(2), stk.H(3)),
        ),
        lambda: stk.Dibromo(
            bromine1=stk.Br(0),
            atom1=stk.C(1),
            bromine2=stk.Br(2),
            atom2=stk.C(3),
            bonders=(stk.C(1),),
            deleters=(stk.Br(0),),
        ),
        lambda: stk.Difluoro(
            fluorine1=stk.F(0),
            atom1=stk.C(1),
            fluorine2=stk.F(2),
            atom2=stk.C(3),
            bonders=(stk.C(1),),
            deleters=(stk.F(0),),
        ),
        lambda: stk.Diol(
            atom1=stk.C(0),
            oxygen1=stk.O(1),
            hydrogen1=stk.H(2),
            atom2=stk.C(3),
            oxygen2=stk.O(4),
            hydrogen2=stk.H(5),
            bonders=(stk.C(0),),
            deleters=(stk.O(1), stk.H(2)),
        ),
        lambda: stk.Fluoro(
            fluorine=stk.F(1),
            atom=stk.C(0),
            bonders=(stk.C(0),),
            deleters=(stk.F(1),),
        ),
        lambda: stk.Iodo(
            iodine=stk.I(0),
            atom=stk.C(1),
            bonders=(stk.C(1),),
            deleters=(stk.I(0),),
        ),
        lambda: stk.PrimaryAmino(
            nitrogen=stk.N(0),
            hydrogen1=stk.H(1),
            hydrogen2=stk.H(2),
            atom=stk.C(3),
            bonders=(stk.C(3),),
            deleters=(stk.N(0), stk.H(1), stk.H(2)),
        ),
        lambda: stk.SecondaryAmino(
            nitrogen=stk.N(0),
            hydrogen=stk.H(1),
            atom1=stk.C(2),
            atom2=stk.C(3),
            bonders=(stk.N(0),),
            deleters=(stk.H(1),),
        ),
        lambda: stk.Thioacid(
            carbon=stk.C(0),
            oxygen=stk.O(1),
            sulfur=stk.S(2),
            hydrogen=stk.H(3),
            atom=stk.C(4),
            bonders=(stk.C(0),),
            deleters=(stk.S(2), stk.H(3)),
        ),
        lambda: stk.Thiol(
            sulfur=stk.S(0),
            hydrogen=stk.H(1),
            atom=stk.C(2),
            bonders=(stk.C(2),),
            deleters=(stk.S(0), stk.H(1)),
        ),
        lambda: stk.SingleAtom(atom=stk.Fe(0)),
    ),
)
def functional_group1(request) -> stk.GenericFunctionalGroup:
    """
    A :class:`.GenericFunctionalGroup` with 1 bonder atom.

    """

    return request.param()


@pytest.fixture
def functional_group1_2(functional_group1):
    """
    A :class:`.GenericFunctionalGroup` with 1 bonder atom.

    """

    return functional_group1


@pytest.fixture(
    params=(
        lambda: stk.Dibromo(
            bromine1=stk.Br(0),
            atom1=stk.C(1),
            bromine2=stk.Br(2),
            atom2=stk.C(3),
            bonders=(stk.C(1), stk.C(2)),
            deleters=(stk.Br(0), stk.Br(2)),
        ),
        lambda: stk.Difluoro(
            fluorine1=stk.F(0),
            atom1=stk.C(1),
            fluorine2=stk.F(2),
            atom2=stk.C(3),
            bonders=(stk.C(1), stk.C(3)),
            deleters=(stk.F(0), stk.F(2)),
        ),
        lambda: stk.Diol(
            atom1=stk.C(0),
            oxygen1=stk.O(1),
            hydrogen1=stk.H(2),
            atom2=stk.C(3),
            oxygen2=stk.O(4),
            hydrogen2=stk.H(5),
            bonders=(stk.C(0), stk.C(3)),
            deleters=(stk.O(1), stk.H(2), stk.O(4), stk.H(5)),
        ),
    ),
)
def functional_group2(request):
    """
    A :class:`.GenericFunctionalGroup` with 2 bonder atoms.

    """

    return request.param()


@pytest.fixture(
    params=(
        (0, 0, 0),
        (-1, 0, 1),
        (-1, -1, -2),
        (1, 2, 3),
    ),
)
def periodicity(request):
    """
    The periodicity of a bond created by a :class:`.Reaction`.

    """

    return request.param


@pytest.fixture(
    params=(1, 2, 9),
)
def bond_order(request):
    """
    The bond order of a bond created by a :class:`.Reaction`.

    """

    return request.param


@pytest.fixture(
    params=(
        lazy_fixture("one_one_reaction"),
        lazy_fixture("one_two_reaction"),
        lazy_fixture("two_two_reaction"),
        lazy_fixture("dative_reaction"),
    ),
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param



import typing
from collections.abc import Callable
from typing import Any

from .fitness_normalizer import FitnessNormalizer

T = typing.TypeVar("T")


class ReplaceFitness(FitnessNormalizer[T]):
    """
    Replaces fitness values of a certain value with a new value.

    Examples:
        *Giving a Fitness Value to Failed Calculations*

        You want to replace fitness values which are ``None`` with
        half the minimum fitness value in the population. A fitness value
        may be ``None`` because the fitness calculation failed for some
        reason.

        .. testcode:: giving-a-fitness-value-to-failed-calculations

            import stk

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: 2,
                record2: None,
            }

            def get_minimum_fitness_value(fitness_values):
                return min(
                    value for value in fitness_values.values()
                    if value is not None
                )

            replacer = stk.ReplaceFitness(
                # The replacement is half the minimum fitness value in the
                # population.
                get_replacement=lambda fitness_values:
                    get_minimum_fitness_value(fitness_values) / 2,
                # Only replace fitness values which are None.
                filter=lambda fitness_values, record:
                    fitness_values[record] is None,
            )

            # Use the replacer.
            normalized_fitness_values = replacer.normalize(fitness_values)
            assert normalized_fitness_values[record1] == 2
            assert normalized_fitness_values[record2] == 1
    """

    def __init__(
        self,
        get_replacement: Callable[[dict[T, Any]], Any],
        filter: Callable[[dict[T, Any], T], bool] = lambda fitness_values,
        record: True,
    ) -> None:
        """
        Parameters:

            get_replacement:
                Calculates the replacement fitness value. The
                entire population passed to :meth:`.normalize`
                is passed to this parameter, regardless
                of what is passed to the `filter` parameter.

            filter:
                A function which returns ``True`` or ``False``. Only
                molecules which return ``True`` will have fitness values
                normalized. By default, all molecules will have fitness
                values normalized.
                The instance passed to the `fitness_values` argument of
                :meth:`.normalize` is passed as the first argument, while
                the second argument will be passed every
                :class:`.MoleculeRecord` in it, one at a time.
        """
        self._get_replacement = get_replacement
        self._filter = filter

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        replacement = self._get_replacement(fitness_values)
        return {
            record: (
                replacement
                if self._filter(fitness_values, record)
                else fitness_value
            )
            for record, fitness_value in fitness_values.items()
        }



"""
Bond Batch
==========

"""

from __future__ import annotations

from typing import Iterable

from stk._internal.bond import Bond
from stk._internal.bond_info import BondInfo
from stk._internal.building_block import BuildingBlock


class _BondBatch:
    """
    A batch of bonds.

    """

    __slots__ = ["_bonds", "_bond_infos"]

    _bonds: list[Bond]
    _bond_infos: list[BondInfo]

    def __init__(
        self,
        bonds: Iterable[Bond],
        id_map: dict[int, int],
        building_block: BuildingBlock,
        building_block_id: int,
    ) -> None:
        """
        Initialize a :class:`.BondBatch` instance.

        Parameters:

            bonds:
                The bonds, which should be added to the batch.

            id_map:
                Maps the ids of atoms held by `bonds`, to the new
                atoms, which the bonds in the :class:`.BondBatch`
                instance should hold.

            building_block:
                The building block from which the bonds originate.

            building_block_id:
                An id, unique to that building block and placement.

        """

        self._bonds = []
        self._bond_infos = []

        for bond in bonds:
            self._bonds.append(bond.with_ids(id_map))
            self._bond_infos.append(
                BondInfo(
                    bond=self._bonds[-1],
                    building_block=building_block,
                    building_block_id=building_block_id,
                )
            )

    def get_bonds(self) -> Iterable[Bond]:
        """
        Yield the bonds in the batch.

        Yields:

            A bond in the batch.

        """

        yield from self._bonds

    def get_bond_infos(self) -> Iterable[BondInfo]:
        """
        Yield info about the bonds in the batch.

        Yields:

            Info about a bond in the batch.

        """

        yield from self._bond_infos



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            path="NCCN.mae",
        ),
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            path="NCCN.xyz",
        ),
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            path="NCCN.pdb",
        ),
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            path="NCCN_with_cell.pdb",
        ),
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            path="NCCN.mol",
        ),
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            path="NCCN.coord",
        ),
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            path="NCCN_with_cell1.coord",
        ),
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            path="NCCN_with_cell2.coord",
        ),
    ),
)
def case_data(request) -> CaseData:
    """
    A :class:`.CaseData` instance.

    """

    return request.param()



def check_bonds(old_state, new_state, building_blocks):
    old_bonds = tuple(old_state.get_bonds())
    new_bonds = tuple(new_state.get_bonds())
    num_added = len(new_bonds) - len(old_bonds)
    assert num_added == sum(bb.get_num_bonds() for bb in building_blocks)
    for bond1, bond2 in zip(
        new_bonds[len(old_bonds) :],
        get_expected_bonds(old_state, building_blocks),
    ):
        is_equivalent_bond(bond1, bond2)


def get_expected_bonds(old_state, building_blocks):
    start_atom_id = sum(1 for _ in old_state.get_atoms())
    for building_block in building_blocks:
        atom_map = {
            atom.get_id(): atom.with_id(id_)
            for id_, atom in enumerate(
                building_block.get_atoms(),
                start_atom_id,
            )
        }
        start_atom_id += building_block.get_num_atoms()
        for bond in building_block.get_bonds():
            yield bond.with_atoms(atom_map)


def is_equivalent_bond(bond1, bond2):
    is_equivalent_atom(bond1.get_atom1(), bond2.get_atom1())
    is_equivalent_atom(bond1.get_atom2(), bond2.get_atom2())


def is_equivalent_atom(atom1, atom2):
    assert atom1.__class__ is atom2.__class__
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_charge() == atom2.get_charge()



import numpy as np
import pytest

import stk

from ....case_data import CaseData
from ...building_blocks import get_linker, get_pd_atom


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.cage.M2L4Lantern(
                    building_blocks={
                        get_pd_atom(): range(2),
                        get_linker(): range(2, 6),
                    },
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C([H])C2C3=C([H])C([H])=[N](->[Pd+2]45<-[N]6=C([H])C("
                "[H])=C(C7=C([H])C(C8=C([H])C([H])=[N](->[Pd+2](<-[N]9=C([H]"
                ")C([H])=C(C(=C1[H])C=2[H])C([H])=C9[H])(<-[N]1=C([H])C([H])"
                "=C(C2=C([H])C([H])=C([H])C(C9=C([H])C([H])=[N]->4C([H])=C9["
                "H])=C2[H])C([H])=C1[H])<-[N]1=C([H])C([H])=C(C2=C([H])C([H]"
                ")=C([H])C(C4=C([H])C([H])=[N]->5C([H])=C4[H])=C2[H])C([H])="
                "C1[H])C([H])=C8[H])=C([H])C([H])=C7[H])C([H])=C6[H])C([H])="
                "C3[H]"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.cage.M2L4Lantern(
                    building_blocks={
                        get_pd_atom(): range(2),
                        get_linker(): range(2, 6),
                    },
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                    vertex_positions={0: np.array([2, 0, 2])},
                ),
            ),
            smiles=(
                "[H]C1=C([H])C2C3=C([H])C([H])=[N](->[Pd+2]45<-[N]6=C([H])C("
                "[H])=C(C7=C([H])C(C8=C([H])C([H])=[N](->[Pd+2](<-[N]9=C([H]"
                ")C([H])=C(C(=C1[H])C=2[H])C([H])=C9[H])(<-[N]1=C([H])C([H])"
                "=C(C2=C([H])C([H])=C([H])C(C9=C([H])C([H])=[N]->4C([H])=C9["
                "H])=C2[H])C([H])=C1[H])<-[N]1=C([H])C([H])=C(C2=C([H])C([H]"
                ")=C([H])C(C4=C([H])C([H])=[N]->5C([H])=C4[H])=C2[H])C([H])="
                "C1[H])C([H])=C8[H])=C([H])C([H])=C7[H])C([H])=C6[H])C([H])="
                "C3[H]"
            ),
            name=name,
        ),
    ),
)
def metal_cage_m2l4_lantern(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



"""
Carboxylic Acid Factory
=======================

"""

from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.utilities import get_atom_ids
from stk._internal.functional_groups.carboxylic_acid import CarboxylicAcid


class CarboxylicAcidFactory(FunctionalGroupFactory):
    """
    Creates :class:`.CarboxylicAcid` instances.

    Creates functional groups from substructures, which match the
    ``[*][C](=[O])[O][H]`` functional group string.

    Examples
    --------
    *Creating Functional Groups with the Factory*

    You want to create a building block which has
    :class:`.CarboxylicAcid` functional groups. You want the carbon
    atom in those functional
    groups to be the *bonder* atom, and the OH group to be a leaving
    group.

    .. testcode:: creating-functional-groups-with-the-factory

        import stk

        building_block = stk.BuildingBlock(
            smiles='OC(=O)CC(=O)O',
            functional_groups=(stk.CarboxylicAcidFactory(), ),
        )

    .. testcode:: creating-functional-groups-with-the-factory
        :hide:

        assert all(
            isinstance(functional_group, stk.CarboxylicAcid)
            for functional_group
            in building_block.get_functional_groups()
        )
        assert building_block.get_num_functional_groups() == 2

    *Changing the Bonder and Deleter Atoms*

    You want to create a building block which has
    :class:`.CarboxylicAcid` functional groups. You want the carbon
    atom to be the *bonder* atom and the oxygen atom to be the
    *deleter* atom.

    .. testcode:: changing-the-bonder-and-deleter-atoms

        import stk

        carboxylic_acid_factory = stk.CarboxylicAcidFactory(
            # The index of the carbon atom in the functional
            # group string (see docstring) is 1.
            bonders=(1, ),
            # The index of the oxygen atom in the functional
            # group string (see docstring) is 2.
            deleters=(2, ),
        )
        building_block = stk.BuildingBlock(
            smiles='OC(=O)CC(=O)O',
            functional_groups=(carboxylic_acid_factory, ),
        )

    .. testcode:: changing-the-bonder-and-deleter-atoms
        :hide:

        fg1, fg2 = building_block.get_functional_groups()
        assert fg1.get_num_bonders() == 1
        assert sum(1 for _ in fg1.get_deleters()) == 1
        assert fg2.get_num_bonders() == 1
        assert sum(1 for _ in fg2.get_deleters()) == 1

        assert all(
            isinstance(atom, stk.C)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_bonders()
        )
        assert all(
            isinstance(atom, stk.O)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_deleters()
        )


    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, bonders=(1,), deleters=(3, 4), placers=None):
        """
        Initialize a :class:`.CarboxylicAcidFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        for atom_ids in get_atom_ids("[*][C](=[O])[O][H]", molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield CarboxylicAcid(
                carbon=atoms[1],
                oxygen1=atoms[2],
                oxygen2=atoms[3],
                hydrogen=atoms[4],
                atom=atoms[0],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



from . import vertices  # noqa
from .metal_complex import *  # noqa
from .octahedral import *  # noqa
from .paddlewheel import *  # noqa
from .porphyrin import *  # noqa
from .square_planar import *  # noqa



import typing
from collections import abc

from stk._internal.databases.value import ValueDatabase
from stk._internal.ea.molecule_record import MoleculeRecord

from .fitness_calculator import FitnessCalculator

T = typing.TypeVar("T", bound=MoleculeRecord)


class FitnessFunction(FitnessCalculator[T]):
    """
    Takes a function and uses it as a fitness calculator.

    Examples:

        *Calculating Fitness Values*

        .. testcode:: calculating-fitness-values

            import stk

            # First create a function to use as the fitness function. For
            # example, assume that the fitness value is given by the number
            # of atoms.
            def get_num_atoms(record):
                return record.get_molecule().get_num_atoms()

            # Then use it to create a fitness calculator.
            fitness_calculator = stk.FitnessFunction(
                fitness_function=get_num_atoms,
            )

            # Use the fitness calculator to get fitness values.
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(stk.BuildingBlock('BrCCBr'), ),
                    repeating_unit='A',
                    num_repeating_units=1,
                ),
            )
            value1 = fitness_calculator.get_fitness_value(record)

        .. testcode:: calculating-fitness-values
            :hide:

            assert value1 == stk.BuildingBlock('BrCCBr').get_num_atoms()

        *Storing Fitness Values in a Database*

        Sometimes you want to store fitness values in a database, you
        can do this by providing the `output_database` parameter.

        .. testsetup:: storing-fitness-values-in-a-database

            import stk

            # Change the database used, so that when a developer
            # runs the doctests locally, their "stk" database is not
            # contaminated.
            _test_database = '_stk_doctest_database'
            _old_init = stk.ValueMongoDb
            stk.ValueMongoDb = lambda mongo_client, collection: (
                _old_init(
                    mongo_client=mongo_client,
                    database=_test_database,
                    collection=collection,
                )
            )

            # Change the database MongoClient will connect to.

            import os
            import pymongo

            _mongo_client = pymongo.MongoClient
            _mongodb_uri = os.environ.get(
                'MONGODB_URI',
                'mongodb://localhost:27017/'
            )
            pymongo.MongoClient = lambda: _mongo_client(_mongodb_uri)

        .. testcode:: storing-fitness-values-in-a-database

            import stk
            import pymongo

            # Create a database which stores the fitness value of each
            # molecule.
            fitness_db = stk.ValueMongoDb(
                # This connects to a local database - so make sure you have
                # local MongoDB server running. You can also connect to
                # a remote MongoDB with MongoClient(), read to pymongo
                # docs to see how to do that.
                mongo_client=pymongo.MongoClient(),
                collection='fitness_values',
            )

            # Define a fitness function.
            def get_num_atoms(record):
                return record.get_molecule().get_num_atoms()

            # Create a fitness calculator.
            fitness_calculator = stk.FitnessFunction(
                fitness_function=get_num_atoms,
                output_database=fitness_db,
            )

            # Calculate fitness values.
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(stk.BuildingBlock('BrCCBr'), ),
                    repeating_unit='A',
                    num_repeating_units=1,
                ),
            )
            value = fitness_calculator.get_fitness_value(record)

            # You can retrieve the fitness values from the database.
            value1 = fitness_db.get(stk.BuildingBlock('BrCCBr'))

        .. testcode:: storing-fitness-values-in-a-database
            :hide:

            assert value == value1

        .. testcleanup:: storing-fitness-values-in-a-database

            stk.ValueMongoDb = _old_init
            pymongo.MongoClient().drop_database(_test_database)
            pymongo.MongoClient = _mongo_client

        *Caching Fitness Values*

        Usually, if you calculate the fitness value of a molecule, you
        do not want to re-calculate it, because this may be expensive,
        and the fitness value is going to be the
        same anyway. By using the `input_database` parameter, together
        with the `output_database` parameter, you can make sure you store
        and retrieve calculated fitness values instead of repeating the
        same calculation multiple times.

        The `input_database` is checked before a calculation happens, to
        see if the value already exists, while the `output_database` has
        the calculated fitness value deposited into it.

        .. testsetup:: caching-fitness-values

            import stk

            # Change the database used, so that when a developer
            # runs the doctests locally, their "stk" database is not
            # contaminated.
            _test_database = '_stk_doctest_database'
            _old_init = stk.ValueMongoDb
            stk.ValueMongoDb = lambda mongo_client, collection: (
                _old_init(
                    mongo_client=mongo_client,
                    database=_test_database,
                    collection=collection,
                )
            )

            # Change the database MongoClient will connect to.

            import os
            import pymongo

            _mongo_client = pymongo.MongoClient
            _mongodb_uri = os.environ.get(
                'MONGODB_URI',
                'mongodb://localhost:27017/'
            )
            pymongo.MongoClient = lambda: _mongo_client(_mongodb_uri)

        .. testcode:: caching-fitness-values

            import stk
            import pymongo

            # You can use the same database for both the input_database
            # and output_database parameters.
            fitness_db = stk.ValueMongoDb(
                # This connects to a local database - so make sure you have
                # local MongoDB server running. You can also connect to
                # a remote MongoDB with MongoClient(), read to pymongo
                # docs to see how to do that.
                mongo_client=pymongo.MongoClient(),
                collection='fitness_values',
            )

            # Define a fitness function.
            def get_num_atoms(record):
                return record.get_molecule().get_num_atoms()

            # Create a fitness calculator.
            fitness_calculator = stk.FitnessFunction(
                fitness_function=get_num_atoms,
                input_database=fitness_db,
                output_database=fitness_db,
            )

            # Assuming that a fitness value for this molecule was not
            # deposited into the database in a previous session, this
            # will calculate the fitness value.
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(stk.BuildingBlock('BrCCBr'), ),
                    repeating_unit='A',
                    num_repeating_units=1,
                ),
            )
            value1 = fitness_calculator.get_fitness_value(record)
            # This will not re-calculate the fitness value, instead,
            # value1 will be retrieved from the database.
            value2 = fitness_calculator.get_fitness_value(record)

        .. testcode:: caching-fitness-values
            :hide:

            assert value1 == value2

        .. testcleanup:: caching-fitness-values

            stk.ValueMongoDb = _old_init
            pymongo.MongoClient().drop_database(_test_database)
            pymongo.MongoClient = _mongo_client
    """

    def __init__(
        self,
        fitness_function: abc.Callable[[T], typing.Any],
        input_database: ValueDatabase | None = None,
        output_database: ValueDatabase | None = None,
    ) -> None:
        """
        Parameters:
            fitness_function:
                Takes a single parameter, the molecule
                whose fitness needs to be calculated, and returns its
                fitness value.

            input_database:
                A database to check before calling `fitness_function`. If a
                fitness value exists for a molecule in the database, the
                stored value is returned, instead of calling
                `fitness_function`.

            output_database:
                A database into which the calculate fitness value is
                placed.
        """
        self._fitness_function = fitness_function
        self._input_database = input_database
        self._output_database = output_database

    def get_fitness_value(self, record: T) -> typing.Any:
        if self._input_database is not None:
            try:
                fitness_value = self._input_database.get(
                    molecule=record.get_molecule(),
                )
            except KeyError:
                fitness_value = self._fitness_function(record)
        else:
            fitness_value = self._fitness_function(record)

        if self._output_database is not None:
            self._output_database.put(record.get_molecule(), fitness_value)

        return fitness_value



from __future__ import annotations

import stk


class CaseData:
    """
    A test case.

    Attributes:

        database: The database to test.

        expected_molecules: The expected molecules to get from the
            databases using their smiles as the key.

    """

    def __init__(
        self,
        database: stk.ConstructedMoleculeDatabase,
        expected_molecules: dict[str, stk.ConstructedMolecule],
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters:

            database: The database to test.

            expected_molecules: The expected molecules to get from the
                databases using their smiles as the key.

        """

        self.database = database
        self.expected_molecules = dict(expected_molecules)



from __future__ import annotations

import typing
import warnings
from collections import abc

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.optimizers.null import (
    NullOptimizer,
)
from stk._internal.optimizers.optimizer import (
    Optimizer,
)
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import ReactionFactory
from stk._internal.topology_graphs.edge import (
    Edge,
)
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import (
    Vertex,
)

from .vertices import CycleVertex


class Macrocycle(TopologyGraph):
    """
    Represents a macrocycle topology graph.

    Building blocks with two functional groups are required
    for this topology.

    Examples:

        *Construction*

        This topology graph essentially makes a polymer chain and joins
        the ends, hence the constructor parameters allows you to
        specify the chain

        .. testcode:: construction

            import stk

            macrocycle = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            macrocycle = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        macrocycle.get_atoms(),
                        macrocycle.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in macrocycle.get_bonds()
                ),
            )

        *Suggested Optimization*

        For :class:`.Macrocycle` topologies, it is recommended to use
        the :class:`.MCHammer` optimizer.

        .. testcode:: suggested-optimization

            import stk

            macrocycle = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    optimizer=stk.MCHammer(),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            macrocycle = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    optimizer=stk.MCHammer(),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        macrocycle.get_atoms(),
                        macrocycle.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in macrocycle.get_bonds()
                ),
            )

        *Defining the Orientation of Each Building Block*

        The `orientations` parameter allows the direction of each
        building block along to the chain to be flipped

        .. testcode:: defining-the-orientation-of-each-building-block

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCOCCBr', [stk.BromoFactory()])

            c1 = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(bb1, bb2),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    orientations=(1, 0.5),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCOCCBr', [stk.BromoFactory()])

            macrocycle = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(bb1, bb2),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    orientations=(1, 0.5),
                    random_seed=1,
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        macrocycle.get_atoms(),
                        macrocycle.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in macrocycle.get_bonds()
                ),
            )

        In the above example, ``bb1`` is guaranteed to be flipped,
        ``bb2`` has a 50% chance of being flipped, each time it is
        placed on a node.

        Note that whether a building block will be flipped or not
        is decided during the initialization of :class:`.Macrocycle`

        .. testcode:: defining-the-orientation-of-each-building-block

            # cycle will always construct the same macrocycle.
            cycle = stk.macrocycle.Macrocycle(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(0.65, 0.45),
            )
            # c2 and c3 are guaranteed to be the same as they used the
            # same topology graph.
            c2 = stk.ConstructedMolecule(cycle)
            c3 = stk.ConstructedMolecule(cycle)

            # cycle2 may lead to a different polymer than chain,
            # despite being initialized with the same parameters.
            cycle2 = stk.macrocycle.Macrocycle(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(0.65, 0.45)
            )

            # c4 and c5 are guaranteed to be the same because they used
            # the same topology graph. However, they may be different
            # to c2 and c3.
            c4 = stk.ConstructedMolecule(cycle2)
            c5 = stk.ConstructedMolecule(cycle2)

        The `random_seed` parameter can be used to get reproducible
        results

        .. testcode:: defining-the-orientation-of-each-building-block

            # c6 and c7 are guaranteed to be the same, because cycle3
            # and cycle4 used the same random seed.

            cycle3 = stk.macrocycle.Macrocycle(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(0.65, 0.45),
                random_seed=4,
            )
            c6 = stk.ConstructedMolecule(cycle3)

            cycle4 = stk.macrocycle.Macrocycle(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(0.65, 0.45),
                random_seed=4,
            )
            c7 = stk.ConstructedMolecule(cycle4)

        *Using Numbers to Define the Repeating Unit*

        The repeating unit can also be specified through the indices of
        the building blocks

        .. testcode:: using-numbers-to-define-the-repeating-unit

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCNCBr', [stk.BromoFactory()])
            bb3 = stk.BuildingBlock('BrCNNCBr', [stk.BromoFactory()])

            # c1 and c2 are different ways to write the same thing.
            c1 = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(bb1, bb2, bb3),
                    repeating_unit='ACB',
                    num_repeating_units=3
                )
            )
            c2 = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(bb1, bb2, bb3),
                    repeating_unit=(0, 2, 1),
                    num_repeating_units=3,
                )
            )

    """

    def __init__(
        self,
        building_blocks: tuple[BuildingBlock, ...],
        repeating_unit: typing.Union[str, tuple[int, ...]],
        num_repeating_units: int,
        orientations: tuple[float, ...] | None = None,
        random_seed: int | np.random.Generator | None = None,
        reaction_factory: ReactionFactory = GenericReactionFactory(),
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
        scale_multiplier: float = 1.0,
    ) -> None:
        """
        Initialize a :class:`Macrocycle` instance.

        Parameters:
            building_blocks:
                The building blocks of the macrocycle.

            repeating_unit:
                A string specifying the repeating unit of the
                macrocycle. For example, ``'AB'`` or ``'ABB'``. The
                first building block passed to `building_blocks` is
                ``'A'`` and so on.

                The repeating unit can also be specified by the
                indices of `building_blocks`, for example ``'ABB'``
                can be written as ``(0, 1, 1)``.

            num_repeating_units:
                The number of repeating units which are used to make
                the macrocycle.

            orientations:
                For each character in the repeating unit, a value
                between ``0`` and ``1`` (both inclusive) must be given
                in a :class:`tuple`. It indicates the probability that
                each monomer will have its orientation along the chain
                flipped. If ``0`` then the monomer is guaranteed not
                to flip. If ``1`` it is guaranteed to flip. This
                allows the user to create head-to-head or head-to-tail
                chains, as well as chain with a preference for
                head-to-head or head-to-tail if a number between
                ``0`` and ``1`` is chosen. If ``None``, then ``0`` is
                picked in every case.

                It is also possible to supply an orientation for every
                vertex in the final topology graph. In this case, the
                length of `orientations` must be equal to
                ``len(repeating_unit)*num_repeating_units``.

            random_seed:
                The random seed to use when choosing random
                orientations.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

            scale_multiplier:
                Scales the positions of the vertices.

        Raises:

            :class:`ValueError`
                If the length of `orientations` is not equal in length
                to `repeating_unit` or to the total number of vertices.

        """

        if orientations is None:
            orientations = tuple(0.0 for i in range(len(repeating_unit)))

        if len(orientations) == len(repeating_unit):
            orientations = orientations * num_repeating_units

        chain_length = len(repeating_unit) * num_repeating_units
        if chain_length == 2:
            warnings.warn(
                "The orientation of macrocycles with chain length "
                f"{chain_length} is not expected to provide robust "
                "alignment and bonding."
            )

        if len(orientations) != chain_length:
            raise ValueError(
                "The length of orientations must match either "
                "the length of repeating_unit or the "
                "total number of vertices."
            )

        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        # Keep these for __repr__.
        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit
        )
        self._num_repeating_units = num_repeating_units

        # Each monomer in the macrocycle is separated by angle_diff.
        angle_diff = (2 * np.pi) / chain_length
        vertices: list[CycleVertex] = []
        edges: list[Edge] = []
        choices = [True, False]
        for vertex_id, flip_chance in enumerate(orientations):
            theta = vertex_id * angle_diff
            vertices.append(
                CycleVertex(
                    id=vertex_id,
                    position=(np.cos(theta), np.sin(theta), 0),
                    flip=random_seed.choice(
                        choices,
                        p=[flip_chance, 1 - flip_chance],
                    ),
                    angle=theta,
                )
            )

            if vertex_id > 0:
                edges.append(
                    Edge(
                        id=len(edges),
                        vertex1=vertices[vertex_id - 1],
                        vertex2=vertices[vertex_id],
                    )
                )

        # Save the chosen orientations for __repr__.
        self._orientations = tuple(
            int(vertex.get_flip()) for vertex in vertices
        )

        edges.append(Edge(len(edges), vertices[0], vertices[-1]))
        super().__init__(
            building_block_vertices=self._get_building_block_vertices(
                building_blocks=building_blocks,
                vertices=vertices,
            ),
            edges=tuple(edges),
            reaction_factory=reaction_factory,
            construction_stages=(),
            num_processes=num_processes,
            optimizer=optimizer,
            edge_groups=None,
            # Handles the different scaling applied to macrocycles.
            scale_multiplier=(
                scale_multiplier
                * len(self._repeating_unit)
                * self._num_repeating_units
            ),
        )

    def clone(self) -> Macrocycle:
        clone = self._clone()
        clone._repeating_unit = self._repeating_unit
        clone._num_repeating_units = self._num_repeating_units
        clone._orientations = self._orientations
        return clone

    @staticmethod
    def _normalize_repeating_unit(
        repeating_unit: typing.Union[str, tuple[int, ...]],
    ) -> tuple[int, ...]:
        if isinstance(repeating_unit, tuple):
            return repeating_unit
        base = ord("A")
        return tuple(ord(letter) - base for letter in repeating_unit)

    def _get_building_block_vertices(
        self,
        building_blocks: tuple[BuildingBlock, ...],
        vertices: abc.Sequence[CycleVertex],
    ) -> dict[BuildingBlock, abc.Sequence[Vertex]]:
        polymer = self._repeating_unit * self._num_repeating_units
        building_block_vertices: dict[BuildingBlock, list[Vertex]]
        building_block_vertices = {}
        for bb_index, vertex in zip(polymer, vertices):
            bb = building_blocks[bb_index]
            building_block_vertices[bb] = building_block_vertices.get(bb, [])
            building_block_vertices[bb].append(vertex)
        return typing.cast(
            dict[BuildingBlock, abc.Sequence[Vertex]],
            building_block_vertices,
        )

    @staticmethod
    def _get_scale(
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
        scale_multiplier: float,
    ) -> float:
        return (
            0.25
            * max(bb.get_maximum_diameter() for bb in building_block_vertices)
            * scale_multiplier
        )

    def with_building_blocks(
        self,
        building_block_map: dict[BuildingBlock, BuildingBlock],
    ) -> Macrocycle:
        return super()._with_building_blocks(building_block_map)

    def __repr__(self) -> str:
        return (
            f"macrocycle.Macrocycle({self._repeating_unit!r}, "
            f"{self._num_repeating_units!r}, "
            f"{self._orientations!r})"
        )



import itertools as it

from ..utilities import is_equivalent_functional_group

import stk
from collections.abc import Sequence
from .case_data import CaseData


def test_get_functional_groups(case_data: CaseData) -> None:
    """
    Test :meth:`.BuildingBlock.get_functional_groups`.

    Parameters:
        case_data:
            A test case. Holds the building block to test and the correct
            functional groups.

    """

    _test_get_functional_groups(
        building_block=case_data.building_block,
        functional_groups=case_data.functional_groups,
    )


def _test_get_functional_groups(
    building_block: stk.BuildingBlock,
    functional_groups: Sequence[stk.FunctionalGroup],
) -> None:
    """
    Test :meth:`.BuildingBlock.get_functional_groups`.

    Parameters:
        building_block:
            The building block to test.

        functional_groups:
            The correct functional groups.

    """

    for fg1, fg2 in it.zip_longest(
        building_block.get_functional_groups(),
        functional_groups,
    ):
        is_equivalent_functional_group(fg1, fg2)



from ..utilities import (
    are_equivalent_functional_groups,
    has_same_structure,
    is_equivalent_building_block,
    is_equivalent_molecule,
)
import stk
from collections.abc import Sequence, Callable, Iterator


def test_with_functional_groups(
    building_block: stk.BuildingBlock,
    get_functional_groups: Callable[
        [stk.BuildingBlock], Iterator[stk.FunctionalGroup]
    ],
) -> None:
    """
    Test :meth:`.BuildingBlock.with_functional_groups`.

    Parameters:
        building_block:
            The building block to test.

        get_functional_groups:
            Takes a single parameter, `building_block` and returns the
            `functional_groups` parameter to use for this test.

    """

    # Save clone to check immutability.
    clone = building_block.clone()
    _test_with_functional_groups(
        building_block=building_block,
        functional_groups=tuple(get_functional_groups(building_block)),
    )
    is_equivalent_building_block(building_block, clone)
    has_same_structure(building_block, clone)


def _test_with_functional_groups(
    building_block: stk.BuildingBlock,
    functional_groups: Sequence[stk.FunctionalGroup],
) -> None:
    """
    Test :meth:`.BuildingBlock.with_functional_groups`.

    Parameters:
        building_block:
            The building block to test.

        functional_groups:
            The functional groups the new building block should hold.

    """

    new = building_block.with_functional_groups(functional_groups)
    are_equivalent_functional_groups(
        new.get_functional_groups(),
        functional_groups,
    )
    is_equivalent_molecule(building_block, new)
    has_same_structure(building_block, new)



import numpy as np
import pytest

from .utilities import get_plane_normal


@pytest.mark.parametrize(
    argnames=("position_matrix", "atom_ids", "normal"),
    argvalues=(
        (
            np.array(
                [
                    [1.0, 2.0, 0.0],
                    [3.0, 4.0, 0.0],
                    [123.0, 423, 0.0],
                ]
            ),
            (0, 1, 2),
            np.array([0.0, 0.0, 1.0]),
        ),
        (
            np.array(
                [
                    [1.0, 2.0, 0.0],
                    [3.0, 4.0, 0.0],
                    [123.0, 423, 105.0],
                ]
            ),
            (0, 1),
            np.array([0.0, 0.0, 1.0]),
        ),
        (
            np.array(
                [
                    [0.0, 2.0, 3.0],
                    [0.0, 4.0, 1.0],
                    [0.0, 423, 12.0],
                ]
            ),
            (0, 1, 2),
            np.array([1.0, 0.0, 0.0]),
        ),
    ),
)
def test_get_plane_normal_helper(position_matrix, atom_ids, normal):
    """
    Test :func:`.get_plane_normal`.

    Parameters
    ----------
    position_matrix : :class:`numpy.ndarray`
        The position matrix to test.

    atom_ids : :class:`tuple` of :class:`int`
        The ids of atoms to include in the test.

    normal : :class:`numpy.ndarray`
        The correct plane normal.

    Returns
    -------
    None : :class:`NoneType`

    """

    result = get_plane_normal(position_matrix, atom_ids)
    # The normal may be parallel or anti-parallel.
    assert abs(abs(result @ normal) - 1) < 1e-32



import pytest

import stk

from ...case_data import CaseData

# With 1 reaction pair.
cycle1 = stk.ConstructedMolecule(
    topology_graph=stk.macrocycle.Macrocycle(
        building_blocks=(
            stk.BuildingBlock(
                smiles="[Br]CC[Br]",
                functional_groups=[stk.BromoFactory()],
            ),
            stk.BuildingBlock(
                smiles="[Br]C(CCI)C[Br]",
                functional_groups=[stk.BromoFactory()],
            ),
        ),
        repeating_unit="ABAAA",
        num_repeating_units=1,
    ),
)
axle1 = stk.ConstructedMolecule(
    topology_graph=stk.polymer.Linear(
        building_blocks=(
            stk.BuildingBlock("BrCCBr", [stk.BromoFactory()]),
            stk.BuildingBlock("BrCNCBr", [stk.BromoFactory()]),
            stk.BuildingBlock("BrCOCI", [stk.BromoFactory()]),
        ),
        repeating_unit="ABABC",
        num_repeating_units=1,
    )
)
rotaxane1 = stk.ConstructedMolecule(
    topology_graph=stk.rotaxane.NRotaxane(
        axle=stk.BuildingBlock.init_from_molecule(axle1),
        cycles=(stk.BuildingBlock.init_from_molecule(cycle1),),
        repeating_unit="A",
        num_repeating_units=1,
    ),
)

# With 2 reaction pairs.
cycle2 = stk.ConstructedMolecule(
    topology_graph=stk.macrocycle.Macrocycle(
        building_blocks=(
            stk.BuildingBlock(
                smiles="[Br]CC[Br]",
                functional_groups=[stk.BromoFactory()],
            ),
            stk.BuildingBlock(
                smiles="[Br]C(CCI)C[Br]",
                functional_groups=[stk.BromoFactory()],
            ),
        ),
        repeating_unit="ABABA",
        num_repeating_units=1,
    ),
)
axle2 = stk.ConstructedMolecule(
    topology_graph=stk.polymer.Linear(
        building_blocks=(
            stk.BuildingBlock("BrCCBr", [stk.BromoFactory()]),
            stk.BuildingBlock("BrCNCBr", [stk.BromoFactory()]),
            stk.BuildingBlock("BrCOCI", [stk.BromoFactory()]),
        ),
        repeating_unit="CABABC",
        num_repeating_units=1,
    )
)
rotaxane2 = stk.ConstructedMolecule(
    topology_graph=stk.rotaxane.NRotaxane(
        axle=stk.BuildingBlock.init_from_molecule(axle2),
        cycles=(stk.BuildingBlock.init_from_molecule(cycle2),),
        repeating_unit="A",
        num_repeating_units=1,
    ),
)


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.small.InternalReaction(
                    building_block=stk.BuildingBlock.init_from_molecule(
                        molecule=rotaxane1,
                        functional_groups=(stk.IodoFactory(),),
                    ),
                    num_reactions=1,
                ),
            ),
            smiles=(
                "[H]N(C([H])([H])C([H])([H])OC([H])([H])C([H])([H])C([H])([H])"
                "C1([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H]"
                ")C([H])([H])C([H])([H])C([H])([H])C1([H])[H])C([H])([H])C([H]"
                ")([H])C([H])([H])C([H])([H])N([H])C([H])([H])C([H])([H])C([H]"
                ")([H])Br"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.small.InternalReaction(
                    building_block=stk.BuildingBlock.init_from_molecule(
                        molecule=rotaxane2,
                        functional_groups=(stk.IodoFactory(),),
                    ),
                    num_reactions=2,
                ),
            ),
            smiles=(
                "[H]N1C([H])([H])C([H])([H])OC([H])([H])C([H])([H])C([H])([H])"
                "C2([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H]"
                ")C([H])(C([H])([H])C([H])([H])C([H])([H])OC([H])([H])C([H])(["
                "H])C([H])([H])C([H])([H])N([H])C([H])([H])C([H])([H])C([H])(["
                "H])C1([H])[H])C([H])([H])C([H])([H])C2([H])[H]"
            ),
            name=name,
        ),
    ),
)
def small_internal(request: pytest.FixtureRequest) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import stk
from dataclasses import dataclass
from collections.abc import Sequence


@dataclass(slots=True, frozen=True)
class CaseData:
    building_block: stk.BuildingBlock
    functional_groups: Sequence[stk.FunctionalGroup]
    known_repr: str
    core_atom_ids: Sequence[int]
    placer_ids: Sequence[int]



from collections.abc import Iterable

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import ReactionFactory
from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import AngledVertex, LinearVertex


class M3L3Triangle(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(2)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC=NC=C2)C3=CC=NC=C3'
            ),
            functional_groups=stk.SmartsFunctionalGroupFactory(
                smarts='[#6]~[#7X2]~[#6]',
                bonders=(1, ),
                deleters=(),
            ),
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M3L3Triangle(
                corners=bb1,
                linkers=bb2,
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(2)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC=NC=C2)C3=CC=NC=C3'
            ),
            functional_groups=stk.SmartsFunctionalGroupFactory(
                smarts='[#6]~[#7X2]~[#6]',
                bonders=(1, ),
                deleters=(),
            ),
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M3L3Triangle(
                corners=bb1,
                linkers=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Both `corner` and `linker` vertices require building blocks with
    two functional groups for this topology. This class replaces the
    `building_blocks` parameter with the `corner` and `linker`
    parameters.

    See :class:`.Cage` for more details and examples.

    """

    def __init__(
        self,
        corners: (
            Iterable[BuildingBlock] | dict[BuildingBlock, tuple[int, ...]]
        ),
        linkers: (
            Iterable[BuildingBlock] | dict[BuildingBlock, tuple[int, ...]]
        ),
        vertex_alignments: dict[int, int] | None = None,
        vertex_positions: dict[int, np.ndarray] | None = None,
        reaction_factory: ReactionFactory = GenericReactionFactory(),
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
        scale_multiplier: float = 1.0,
    ) -> None:
        """
        Parameters:

            corners:
                Can be a :class:`dict` which maps the
                :class:`.BuildingBlock` instances to the ids of the
                vertices it should be placed on.

                Can also be a :class:`.BuildingBlock` instance, which
                should be placed on all corner vertices on the topology
                graph.

            linkers:
                Can be a :class:`dict` which maps the
                :class:`.BuildingBlock` instances to the ids of the
                vertices it should be placed on.

                Can also be a :class:`.BuildingBlock` instance, which
                should be placed on all linker vertices on the topology
                graph.

            vertex_alignments:
                A mapping from the id of a :class:`.Vertex`
                to an :class:`.Edge` connected to it.
                The :class:`.Edge` is used to align the first
                :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
                placed on that vertex. Only vertices which need to have
                their default edge changed need to be present in the
                :class:`dict`. If ``None`` then the default edge is used
                for each vertex. Changing which :class:`.Edge` is used will
                mean that the topology graph represents different
                structural isomers. The edge is referred to by a number
                between ``0`` (inclusive) and the number of edges the
                vertex is connected to (exclusive).

            vertex_positions:
                A mapping from the id of a :class:`.Vertex` to a custom
                :class:`.BuildingBlock` position. The default vertex
                alignment algorithm is still applied. Only vertices
                which need to have their default position changed need
                to be present in the :class:`dict`. Note that any
                vertices with modified positions will not be scaled like
                the rest of the building block positions and will not
                use neighbor placements in its positioning if requested
                by the default topology. If ``None`` then the default
                placement algorithm is used for each vertex.

            reaction_factory:
                The reaction factory to use for creating bonds between
                building blocks.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

            scale_multiplier:
                Scales the positions of the vertices.

        Raises:

            :class:`AssertionError`
                If the any building block does not have a
                valid number of functional groups.

            :class:`ValueError`
                If the there are multiple building blocks with the
                same number of functional_groups in `building_blocks`,
                and they are not explicitly assigned to vertices. The
                desired placement of building blocks is ambiguous in
                this case.

            :class:`~.cage.UnoccupiedVertexError`
                If a vertex of the cage topology graph does not have a
                building block placed on it.

            :class:`~.cage.OverlyOccupiedVertexError`
                If a vertex of the cage topology graph has more than
                one building block placed on it.

        """

        if isinstance(corners, dict):
            building_blocks = corners
        else:
            building_blocks = {corners: (0, 1, 2)}

        if isinstance(linkers, dict):
            linkers_dict = linkers
        else:
            linkers_dict = {linkers: (3, 4, 5)}

        building_blocks.update(
            (building_block, vertices)
            for building_block, vertices in linkers_dict.items()
        )

        super().__init__(
            building_blocks,
            vertex_alignments=vertex_alignments,
            vertex_positions=vertex_positions,
            reaction_factory=reaction_factory,
            num_processes=num_processes,
            optimizer=optimizer,
            scale_multiplier=scale_multiplier,
        )

    _x = 2 * np.sqrt(3) / 4
    _y = 2
    _angled_vertices = (
        AngledVertex(0, np.array([0, _x, 0])),
        AngledVertex(1, np.array([_y / 2, -_x, 0])),
        AngledVertex(2, np.array([-_y / 2, -_x, 0])),
    )

    _vertex_prototypes = (
        *_angled_vertices,
        LinearVertex.init_at_center(
            id=3,
            vertices=(_angled_vertices[0], _angled_vertices[1]),
        ),
        LinearVertex.init_at_center(
            id=4,
            vertices=(_angled_vertices[1], _angled_vertices[2]),
        ),
        LinearVertex.init_at_center(
            id=5,
            vertices=(_angled_vertices[2], _angled_vertices[0]),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(1, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(2, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(3, _vertex_prototypes[2], _vertex_prototypes[4]),
        Edge(4, _vertex_prototypes[2], _vertex_prototypes[5]),
        Edge(5, _vertex_prototypes[0], _vertex_prototypes[5]),
    )

    _num_windows = 1
    _num_window_types = 1



"""
Ring Amine Factory
==================

"""

from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.utilities import get_atom_ids
from stk._internal.functional_groups.ring_amine import RingAmine


class RingAmineFactory(FunctionalGroupFactory):
    """
    Creates :class:`.RingAmine` functional groups.

    Creates functional groups from substructures, which match the
    ``[N]([H])([H])[#6]~[#6]([H])~[#6R1]`` functional group string.

    """

    # Keep this __init__() method, which does nothing, here for the
    # docstring.
    def __init__(self):
        """
        Initialize a :class:`RingAmineFactory` instance.

        """

        return super().__init__()

    def get_functional_groups(self, molecule):
        ids = get_atom_ids(
            query="[N]([H])([H])[#6]~[#6]([H])~[#6R1]",
            molecule=molecule,
        )
        for atom_ids in ids:
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield RingAmine(
                nitrogen=atoms[0],
                hydrogen1=atoms[1],
                hydrogen2=atoms[2],
                carbon1=atoms[3],
                carbon2=atoms[4],
                hydrogen3=atoms[5],
                carbon3=atoms[6],
            )



import numpy as np

from ....utilities import is_clone


def test_with_structure_from_file(
    molecule,
    get_position_matrix,
    path,
):
    """
    Test :meth:`.Molecule.with_structure_from_file`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    get_position_matrix : :class:`callable
        Takes a single parameter, `molecule`, and returns a valid
        position matrix for this test.

    path : :class:`str`
        A path into which the test structure is written.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a copy of the position matrix, to ensure the original
    # molecule is not modified by the test, because it is meant to be
    # immutable.
    position_matrix = molecule.get_position_matrix()
    _test_with_structure_from_file(
        molecule=molecule,
        get_position_matrix=get_position_matrix,
        path=path,
    )
    assert np.all(
        np.equal(
            position_matrix,
            molecule.get_position_matrix(),
        )
    )


def _test_with_structure_from_file(
    molecule,
    get_position_matrix,
    path,
):
    """
    Test :meth:`.Molecule.with_structure_from_file`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    get_position_matrix : :class:`callable
        Takes a single parameter, `molecule`, and returns a valid
        position matrix for this test.

    path : :class:`str`
        A path into which the test structure is written.

    Returns
    -------
    None : :class:`NoneType`

    """

    position_matrix = get_position_matrix(molecule)
    molecule.with_position_matrix(position_matrix).write(path)
    loaded = molecule.with_structure_from_file(path)
    is_clone(loaded, molecule)
    assert np.allclose(
        a=position_matrix,
        b=loaded.get_position_matrix(),
        # Not very precise because the file does not hold many
        # significant figures.
        atol=1e-4,
    )
